# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Anthropic API message conversion."""

import base64
from typing import Annotated, Any, Callable

import langfun.core as lf
from langfun.core import modalities as lf_modalities


class AnthropicMessageConverter(lf.MessageConverter):
  """Converter between Langfun messages and Anthropic API message format.

  This converter translates `lf.Message` objects into the JSON format required
  by the Anthropic API and vice versa. It handles text and modalities like
  images and PDFs by encoding them in base64 format as expected by Anthropic.
  An optional `chunk_preprocessor` can be provided to modify or filter
  chunks before conversion.
  """

  FORMAT_ID = 'anthropic'

  chunk_preprocessor: Annotated[
      Callable[[str | lf.Modality], Any] | None,
      (
          'Chunk preprocessor for Langfun chunk to Anthropic chunk conversion. '
          'It will be applied before each Langfun chunk is converted. '
          'If it returns None, the chunk will be skipped.'
      )
  ] = None

  def to_value(self, message: lf.Message) -> dict[str, Any]:
    """Converts a Langfun message to Anthropic API."""
    content = []
    for chunk in message.chunk():
      if self.chunk_preprocessor:
        chunk = self.chunk_preprocessor(chunk)
        if chunk is None:
          continue

      if isinstance(chunk, str):
        content.append({'type': 'text', 'text': chunk})
      elif isinstance(chunk, lf_modalities.Mime):
        if isinstance(chunk, lf_modalities.Image):
          content.append(
              dict(
                  type='image',
                  source=dict(
                      type='base64',
                      media_type=chunk.mime_type,
                      data=base64.b64encode(chunk.to_bytes()).decode(),
                  ),
              )
          )
        elif isinstance(chunk, lf_modalities.PDF):
          content.append(
              dict(
                  type='document',
                  source=dict(
                      type='base64',
                      media_type=chunk.mime_type,
                      data=base64.b64encode(chunk.to_bytes()).decode(),
                  ),
              )
          )
        else:
          raise NotImplementedError(
              f'Modality conversion not implemented: {chunk!r}'
          )
    return dict(role=self.get_role(message), content=content)

  def from_value(self, value: dict[str, Any]) -> lf.Message:
    """Returns a Langfun message from Anthropic message."""
    message_cls = self.get_message_cls(
        self._safe_read(value, 'role', default='assistant')
    )
    content = self._safe_read(value, 'content', default=[])
    assert isinstance(content, list)

    chunks = []
    thought_chunks = []
    for part in content:
      t = self._safe_read(part, 'type')
      if t == 'text':
        chunks.append(self._safe_read(part, 'text'))
      elif t == 'thinking':
        thought_chunks.append(self._safe_read(part, 'thinking'))
      elif t in ('image', 'document'):
        source = self._safe_read(part, 'source')
        chunks.append(
            lf_modalities.Mime.class_from_mime_type(
                self._safe_read(source, 'media_type')
            ).from_bytes(base64.b64decode(self._safe_read(source, 'data')))
        )
      elif t in ('server_tool_use', 'web_search_tool_result'):
        continue
      else:
        raise ValueError(f'Unsupported content part: {part!r}.')
    message = message_cls.from_chunks(chunks)
    if thought_chunks:
      message.set('thought', message_cls.from_chunks(thought_chunks))
    return message


def _as_anthropic_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns an Anthropic format message."""
  return AnthropicMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_anthropic_format(
    cls,
    anthropic_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the Anthropic format message."""
  del cls
  return AnthropicMessageConverter(**kwargs).from_value(anthropic_message)

# Set shortcut methods in lf.Message.
lf.Message.as_anthropic_format = _as_anthropic_format
lf.Message.from_anthropic_format = _from_anthropic_format



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for modality (e.g. Image, Video, etc.)."""

import abc
import contextlib
import functools
import hashlib
import re
from typing import Any, ContextManager, Iterator
from langfun.core import component
import pyglove as pg


class Modality(component.Component, pg.views.HtmlTreeView.Extension):
  """Base class for representing non-text content in prompts.

  `lf.Modality` is the base class for multimodal objects such as `lf.Image`,
  `lf.Audio`, and `lf.Video`. It allows these non-text inputs to be
  seamlessly embedded within text prompts for processing by multimodal
  language models.

  When a `Modality` object is rendered within an `lf.Template`, it is
  replaced by a text marker (e.g., `<<[[image:b10a8db1]]>>`), and the
  modality object itself is stored in the `referred_modalities` field of
  the resulting `lf.Message`. This allows language models to associate
  the placeholder with its content during processing.

  **Example:**

  ```python
  import langfun as lf

  image = lf.Image.from_path('/path/to/image.png')
  prompt = lf.Template('What is in this image? {{image}}', image=image)

  message = prompt.render()
  print(message.text)
  # Output: What is in this image? <<[[image:b10a8db1]]>>

  print(message.modalities())
  # Output: [<Image object>]
  ```
  """

  REF_START = '<<[['
  REF_END = ']]>>'

  def _on_bound(self):
    super()._on_bound()
    # Invalidate cached hash and id if modality member is changed.
    self.__dict__.pop('hash', None)
    self.__dict__.pop('id', None)

  def format(self, *args, **kwargs) -> str:
    if not pg.object_utils.thread_local_get(_TLS_MODALITY_AS_REF, False):
      return super().format(*args, **kwargs)

    capture_scope = get_modality_capture_context()
    if capture_scope is not None:
      capture_scope.capture(self)
    return Modality.text_marker(self.id)

  def __str_kwargs__(self) -> dict[str, Any]:
    # For modality objects, we don't want to use markdown format when they
    # are rendered as parts of the prompt.
    kwargs = super().__str_kwargs__()
    kwargs.pop('markdown', None)
    return kwargs

  @abc.abstractmethod
  def to_bytes(self) -> bytes:
    """Returns content in bytes."""

  @functools.cached_property
  def hash(self) -> str:
    """Returns a 8-byte MD5 hash as the identifier for this modality object."""
    return hashlib.md5(self.to_bytes()).hexdigest()[:8]

  @classmethod
  def text_marker(cls, var_name: str) -> str:
    """Returns a marker in the text for this object."""
    return Modality.REF_START + var_name + Modality.REF_END

  @functools.cached_property
  def id(self) -> str | None:
    """Returns the referred name of this object in its template."""
    modality_type = _camel_to_snake(self.__class__.__name__)
    return f'{modality_type}:{self.hash}'

  @classmethod
  def from_value(cls, value: pg.Symbolic) -> dict[str, 'Modality']:
    """Returns a dict of path to modality from a symbolic value."""
    modalities = {}
    def _visit(k, v, p):
      del k, p
      if isinstance(v, Modality):
        modalities[v.id] = v
        return pg.TraverseAction.CONTINUE
      return pg.TraverseAction.ENTER

    pg.traverse(value, _visit)
    return modalities


class ModalityRef(pg.Object, pg.typing.CustomTyping):
  """Lightweight placeholder for a `lf.Modality` object in a symbolic tree.

  `ModalityRef` acts as a reference to a `Modality` object (like `lf.Image`
  or `lf.Audio`) within a structured object hierarchy (e.g., a `pg.Object`).
  Instead of embedding potentially large modality data directly, `ModalityRef`
  stores only the ID of the modality object.

  This is useful in scenarios where structured objects are serialized or
  manipulated, and it's more efficient to refer to modalities by ID rather
  than copying their content. The `lf.ModalityRef.placehold()` class method
  can be used to replace `Modality` instances in a symbolic object with
  `ModalityRef` placeholders, while `lf.ModalityRef.restore()` can reinstate
  the original `Modality` objects using a lookup table.

  **Example:**

  ```python
  import langfun as lf
  import pyglove as pg

  class ImagePair(pg.Object):
    image1: lf.Image
    image2: lf.Image

  pair = ImagePair(
      image1=lf.Image(content=b'abc'), image2=lf.Image(content=b'def')
  )
  modalities = lf.Modality.from_value(pair)

  # Replace Image objects with ModalityRef placeholders
  pair_with_refs = lf.ModalityRef.placehold(pair)
  print(pair_with_refs.image1)
  # Output: ModalityRef(id='image:d81e5a68')

  # Restore Image objects from ModalityRef placeholders
  pair_restored = lf.ModalityRef.restore(pair_with_refs, modalities)
  assert pair_restored.image1.content == b'abc'
  ```
  """

  id: str

  def custom_apply(
      self, path: pg.KeyPath, value_spec: pg.ValueSpec, *args, **kwargs
  ) -> tuple[bool, Any]:
    return (False, self)

  @classmethod
  def placehold(cls, value: pg.Symbolic) -> pg.Symbolic:
    """Returns a copy of value by replacing modality objects with refs.

    Args:
      value: A symbolic value.

    Returns:
      A copy of value with all child `Modality` objects replaced with
        `ModalityRef` objects.
    """

    def _placehold(k, v, p):
      del k, p
      if isinstance(v, Modality):
        return ModalityRef(id=v.id)
      return v
    return value.clone().rebind(_placehold, raise_on_no_change=False)

  @classmethod
  def restore(cls, value: pg.Symbolic, modalities: dict[str, Modality]) -> Any:
    """Returns a copy of value by replacing refs with modality objects."""
    def _restore(k, v, p):
      del k, p
      if isinstance(v, ModalityRef):
        modality_object = modalities.get(v.id)
        if modality_object is None:
          raise ValueError(
              f'Modality {v.id} not found in modalities {modalities.keys()}'
          )
        return modality_object
      return v
    return value.rebind(_restore, raise_on_no_change=False)


class ModalityError(RuntimeError):  # pylint: disable=g-bad-exception-name
  """Exception raised when modality is not supported."""


#
# Context managers to deal with modality objects.
#


_TLS_MODALITY_CAPTURE_SCOPE = '__modality_capture_scope__'
_TLS_MODALITY_AS_REF = '__format_modality_as_ref__'


def format_modality_as_ref(enabled: bool = True) -> ContextManager[None]:
  """A context manager that formats modality objects as references."""
  return pg.object_utils.thread_local_value_scope(
      _TLS_MODALITY_AS_REF, enabled, False
  )


class _ModalityCaptureContext:
  """A context to capture modality objects when being rendered."""

  def __init__(self):
    self._references: dict[str, pg.Ref[Modality]] = {}

  def capture(self, modality: Modality) -> None:
    """Captures the modality object."""
    self._references[modality.id] = pg.Ref(modality)

  @property
  def references(self) -> dict[str, pg.Ref[Modality]]:
    """Returns the modality references captured in this context."""
    return self._references


@contextlib.contextmanager
def capture_rendered_modalities() -> Iterator[dict[str, pg.Ref[Modality]]]:
  """Capture modality objects whose references is being rendered.

  Example:
    ```
    image = lf.Image.from_url(...)
    with lf.modality.capture_rendered_modalities() as rendered_modalities:
      with lf.modality.format_modality_as_ref():
        print(f'Hello {image}')
    self.assertEqual(rendered_modalities, {'image:<hash>': pg.Ref(image)})
    ```
  """
  context = get_modality_capture_context()
  top_level = context is None
  if top_level:
    context = _ModalityCaptureContext()
    pg.object_utils.thread_local_set(_TLS_MODALITY_CAPTURE_SCOPE, context)

  try:
    yield context.references  # pylint: disable=attribute-error
  finally:
    if top_level:
      pg.object_utils.thread_local_del(_TLS_MODALITY_CAPTURE_SCOPE)


def get_modality_capture_context() -> _ModalityCaptureContext | None:
  """Returns the current modality capture context."""
  return pg.object_utils.thread_local_get(_TLS_MODALITY_CAPTURE_SCOPE, None)


def _camel_to_snake(name: str) -> str:
  """Converts a camelCase name to snake_case."""
  return re.sub(
      pattern=r'([A-Z]+)', repl=r'_\1', string=name
  ).lower().lstrip('_')



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common metrics for Langfun evaluation."""


import abc
import collections
import threading
from typing import Annotated, Any

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import metric_values
import pyglove as pg


Rate = metric_values.Rate
Average = metric_values.Average


class Metric(pg.Object, pg.views.HtmlTreeView.Extension):
  """Interface for an evaluation metric.

  A metric is used to evaluate the quality of the outputs produced by an
  evaluation. It works by auditing each processed example via its `audit`
  method, which in turn calls the user-overridable `_audit` method to perform
  metric-specific logic and update metric values. Metrics can compute multiple
  values (e.g., precision, recall, F1 score) which are exposed via the
  `values` method.
  """

  name: Annotated[
      str,
      (
          'Name of the metric, which will be used as the key in the dict '
          'returned by `Experiment.metric_values()`'
      )
  ]

  def _on_bound(self):
    super()._on_bound()
    self._label_group = None
    self._lock = threading.Lock()

  def update(
      self,
      example: example_lib.Example,
      force_recompute: bool = False
  ) -> dict[str, Any]:
    """Updates metric values with a processed example.

    Args:
      example: The processed example.
      force_recompute: Whether to force recompute the metric metadata even if
        they are already present.

    Returns:
      A dict of metric metadata.
    """
    if (force_recompute
        or example.metric_metadata is None
        or self.name not in example.metric_metadata):
      metadata = self.compute_metric_metadata(example)
    else:
      metadata = example.metric_metadata[self.name]
    self.update_metric_values(example.id, metadata)
    self._update_view()
    return metadata

  @abc.abstractmethod
  def compute_metric_metadata(
      self, example: example_lib.Example
  ) -> dict[str, Any]:
    """Subclasses should override this method to implement the metric logic."""

  @abc.abstractmethod
  def update_metric_values(
      self, example_id: int, metric_metadata: dict[str, Any]
  ) -> None:
    """Update metric values based on metric metadata."""

  @abc.abstractmethod
  def values(self) -> list[metric_values.MetricValue]:
    """Returns all the values computed by this metric."""

  def reset(self) -> None:
    """Resets the metric values."""
    for v in self.values():
      v.reset()

  def merge_from(self, other: 'Metric') -> 'Metric':
    """Merges the values from another metric."""
    for v1, v2 in zip(self.values(), other.values()):
      v1.merge_from(v2)
    return self

  def _update_view(self):
    """Refreshes the metric values."""
    if self._label_group is None:
      return

    for label, value in zip(self._label_group.labels, self.values()):
      label.update(
          text=self._metric_value_text(value),
          tooltip=self._metric_value_tooltip(value),
      )

  def _metric_value_text(self, metric_value: metric_values.MetricValue) -> str:
    """Returns the label text for the metric value."""
    return str(metric_value)

  def _metric_value_tooltip(
      self, metric_value: metric_values.MetricValue) -> str:
    """Returns the label text for the metric value."""
    with pg.str_format(verbose=True):
      return f'{metric_value.sym_path.key}: {metric_value}'

  def _metric_label_text(self) -> str:
    return ''.join(
        c for c in self.__class__.__name__
        if c.isalnum() and not c.islower()
    )

  def _metric_label_tooltip(self) -> str:
    return self.__class__.__type_name__

  def _html_tree_view(
      self,
      *,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs,
  ) -> pg.Html:
    """Renders the content of the metric value."""
    extra_flags = extra_flags or {}
    interactive = extra_flags.get('interactive', True)
    label_group = self._label_group
    if label_group is None:
      label_group = pg.views.html.controls.LabelGroup(
          [
              pg.views.html.controls.Label(
                  self._metric_value_text(mv),
                  tooltip=self._metric_value_tooltip(mv),
                  css_classes=[mv.sym_path.key, 'metric-value'],
                  interactive=interactive,
              ) for mv in self.values()
          ],
          name=pg.views.html.controls.Label(
              self._metric_label_text(),
              tooltip=self._metric_label_tooltip(),
              css_classes=[
                  'metric-name',
                  pg.object_utils.camel_to_snake(self.__class__.__name__, '-')
              ],
              interactive=False,
          ),
          css_classes=['metric-container'],
      )
      if interactive:
        self._label_group = label_group
    return label_group.to_html()

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .metric-container {
          display: inline-flex;
          overflow: hidden;
          border-radius: 5px;
          border: 0px;
          margin: 5px;
          padding: 0px;
        }
        .metric-container .label-container {
          vertical-align: middle;
        }
        .metric-value.oop_errors {
          color: magenta;
          background-color: #f9e6eb;
        }
        .metric-value.non_oop_errors {
          color: red;
          background-color: #fdcccc;
        }
        """
    ]

#
# Common metrics.
#


class MetricBase(Metric):
  """Base class for common metrics.

  `MetricBase` provides common functionalities for metrics, such as automatic
  error counting based on whether an example has an error during evaluation.
  It distinguishes between Object-Oriented Programming (OOP) errors
  (e.g. `MappingError` during structured output generation) and other errors.
  Subclasses should implement `_audit_processed` for metric computation on
  successfully processed examples.
  """

  oop_errors: Rate | None = Rate()
  non_oop_errors: Rate | None = Rate()

  def _on_bound(self) -> None:
    super()._on_bound()
    self._error_breakdown = collections.defaultdict(list)

  def reset(self) -> None:
    """Resets the metric."""
    super().reset()
    self._error_breakdown = collections.defaultdict(list)

  def compute_metric_metadata(
      self, example: example_lib.Example
  ) -> dict[str, Any]:
    """Computes the metric metadata for the example."""
    if example.error is None:
      return self._compute_metric_metadata(example)
    return self._compute_metric_metadata_with_processing_error(example)

  def update_metric_values(
      self,
      example_id: int,
      metric_metadata: dict[str, Any]
  ) -> None:
    """Collects the metric metadata."""
    # NOTE(daiyip): the metric values are being updated concurrently, so we
    # uses a lock to avoid race condition. We might consider relaxing the lock
    # later if metric auditing becomes a bottleneck.
    with self._lock:
      for v in self.values():
        v.increment_total()

    if 'error' in metric_metadata:
      self._update_metric_values_with_processing_error(
          example_id, metric_metadata
      )
    else:
      self._update_metric_values(example_id, metric_metadata)

  @abc.abstractmethod
  def _compute_metric_metadata(
      self,
      example: example_lib.Example
  ) -> dict[str, Any]:
    """Computes the metric metadata for the example."""

  def _compute_metric_metadata_with_processing_error(
      self,
      example: example_lib.Example
  ) -> dict[str, Any]:
    """Audits the evaluation example after processing."""
    assert example.error is not None
    return dict(error=example.error.tag)

  @abc.abstractmethod
  def _update_metric_values(self, metadata: dict[str, Any]) -> None:
    """Update metric values based metric metadata."""

  def _update_metric_values_with_processing_error(
      self,
      example_id: int,
      metric_metadata: dict[str, Any]
  ) -> None:
    """Updates metric values with processing error."""
    error_tag = metric_metadata.get('error')
    assert error_tag is not None, (example_id, metric_metadata)
    self._error_breakdown[error_tag].append(example_id)
    if error_tag.startswith('MappingError'):
      self.oop_errors.add(example_id, 1)
    else:
      self.non_oop_errors.add(example_id, 1)
    self._error_breakdown[error_tag].append(example_id)

  def _oop_errors_breakdown(self) -> str | None:
    """Returns the OOP error breakdown as a string."""
    return '\n'.join(
        [
            f'- {k}: {len(v)}' for k, v in self._error_breakdown.items()
            if k.startswith('MappingError')
        ]
    ) or None

  def _non_oop_errors_breakdown(self) -> str | None:
    """Returns the non-OOP error breakdown as a string."""
    return '\n'.join(
        [
            f'- {k}: {len(v)}' for k, v in self._error_breakdown.items()
            if not k.startswith('MappingError')
        ]
    ) or None

  def _sym_nondefault(self) -> dict[str, Any]:
    """Overrides nondefault valuesso volatile values are not included."""
    return dict()


class Match(MetricBase):
  """Metric for matching outputs against ground truth.

  This metric computes match and mismatch rates by comparing the output of
  an example with its ground truth. By default, it looks for a `groundtruth`
  attribute in `example.input` for comparison. Users can customize this behavior
  by subclassing `Match` and overriding the `match` method.
  """

  name = 'match'
  matches: Rate = Rate()
  mismatches: Rate = Rate()

  def match(
      self, example_input: Any, output: Any
  ) -> bool | tuple[bool, dict[str, Any]]:
    """Returns whether the output matches the groundtruth from the example.

    Args:
      example_input: The example input which contains the groundtruth.
      output: The output to match against.

    Returns:
      True if the output matches the groundtruth, False otherwise.
      Or a tuple of (match, metadata).
    """
    groundtruth = getattr(example_input, 'groundtruth', pg.MISSING_VALUE)
    if pg.MISSING_VALUE == groundtruth:
      raise ValueError(
          f'`groundtruth` is not present in the example ({example_input}). '
          'Please subclassing `Match` and override the `match` method to '
          'support custom example format.'
      )
    return pg.eq(output, groundtruth)

  def _compute_metric_metadata(
      self, example: example_lib.Example
  ) -> dict[str, Any]:
    """Computes the metric metadata for the example."""
    metadata = {}
    is_correct = self.match(example.input, example.output)
    if isinstance(is_correct, tuple):
      is_correct, metadata = is_correct

    metadata['is_correct'] = is_correct
    return metadata

  def _update_metric_values(
      self, example_id: int, metadata: dict[str, Any]
  ) -> None:
    """Update metric values based metric metadata."""
    is_correct = metadata.get('is_correct')
    assert is_correct is not None, (example_id, metadata)
    if is_correct:
      self.matches.add(example_id, 1)
    else:
      assert not is_correct
      self.mismatches.add(example_id, 1)

  def values(self) -> list[metric_values.MetricValue]:
    """Returns all the values computed by this metric."""
    return [
        self.matches,
        self.mismatches,
        self.oop_errors,
        self.non_oop_errors
    ]

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .metric-name.match {
          padding: 5px;
          color: white;
          background-color: purple;
        }
        .metric-value.matches {
          color: green;
          background-color: #dcefbe;
        }
        .metric-value.mismatches {
          color: orange;
          background-color: #ffefc4;
        }
        """
    ]


class Score(MetricBase):
  """Base class for scoring metrics.

  `Score` is a base class for metrics that assign a numerical score to each
  example's output (e.g., evaluating quality on a scale of 1-5).
  It automatically computes the average score across all examples.
  Subclasses must implement the `score` method to define how an example
  should be scored.
  """

  name = 'score'
  average_score: Average = Average()

  @abc.abstractmethod
  def score(
      self,
      example_input: Any,
      output: Any) -> float | tuple[float, dict[str, Any]]:
    """Returns the score based on the example and output.

    Args:
      example_input: The example input based on which the output is generated.
      output: The output to score.

    Returns:
      A float score. Or a tuple of (score, metadata).
    """

  def _compute_metric_metadata(
      self, example: example_lib.Example
  ) -> dict[str, Any]:
    """Computes the metric metadata for the example."""
    metadata = {}
    score = self.score(example.input, example.output)
    if isinstance(score, tuple):
      score, metadata = score
    metadata['score'] = score
    return metadata

  def _update_metric_values(
      self, example_id: int, metadata: dict[str, Any]
  ) -> None:
    """Update metric values based metric metadata."""
    score = metadata.get('score')
    assert score is not None, (example_id, metadata)
    self.average_score.add(example_id, score)

  def values(self) -> list[metric_values.MetricValue]:
    """Returns all the values computed by this metric."""
    return [
        self.average_score,
        self.oop_errors,
        self.non_oop_errors
    ]

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .metric-name.score {
          padding: 5px;
          color: white;
          background-color: blue;
        }
        .metric-value.average_score {
          color: blue;
          background-color: #b0c7f6;
        }
        """
    ]



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checkpointing evaluation runs."""
import abc
import datetime
import os
import re
import threading
import traceback
from typing import Annotated

import langfun.core as lf
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
import pyglove as pg

Example = example_lib.Example
Experiment = experiment_lib.Experiment
Runner = experiment_lib.Runner


class Checkpointer(experiment_lib.Plugin):
  """Base class for checkpointing evaluation examples.

  `Checkpointer` is a plugin that saves the state of processed examples
  incrementally during an experiment run, allowing the experiment to be resumed
  later. When an experiment starts, the checkpointer loads any previously saved
  examples from an earlier run (or a warm-start run) into `experiment.state`,
  so the runner can skip processing them again.
  Subclasses should implement `_list_checkpoint_files` to identify
  checkpoint files to load, and `_save_example` to save a newly processed
  example.
  """

  checkpoint_filename: Annotated[
      str,
      'Checkpoint file pattern.'
  ] = 'checkpoint.jsonl'

  enable_inprogress_file: Annotated[
      bool,
      'If True, write file "<example_id>.inprogress" when example gets started.'
  ] = True

  max_ckpt_loading_threads: Annotated[
      int,
      'Max number of workers for loading checkpoint files at startup.'
  ] = 128

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    if not experiment.is_leaf:
      return

    current_run = runner.current_run
    if current_run.reprocess is not True:  # pylint: disable=g-bool-id-comparison
      if current_run.input_root != current_run.output_root:
        experiment.info(
            f'Warm starting from directory: {current_run.input_root}.'
        )
      self._load_experiment(runner, experiment)

    example_ids_to_evaluate = current_run.examples_to_evaluate(experiment)
    if experiment.state.ckpt_examples:
      loaded_example_ids = list(
          sorted(experiment.state.ckpt_examples.keys())
      )
      example_ids_to_evaluate -= set(loaded_example_ids)
      example_ids_to_evaluate = list(sorted(example_ids_to_evaluate))
      experiment.info(
          f'{len(experiment.state.ckpt_examples)} examples '
          'loaded from checkpoint files. Their outputs will be used '
          f'for recomputing metrics. Example IDs: {loaded_example_ids}.'
      )
      experiment.info(
          f'{len(example_ids_to_evaluate)} examples will be processed from '
          f'scratch. Example IDs: {example_ids_to_evaluate}.'
      )
    else:
      experiment.info(
          'No examples are loaded from checkpoint files. '
          f'{len(example_ids_to_evaluate)} examples will be processed from '
          f'scratch. Example IDs: {example_ids_to_evaluate}.'
      )

  def on_example_start(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    if self.enable_inprogress_file:
      def _save_inprogress_file(example: Example):
        inprogress_file = runner.current_run.output_path_for(
            experiment, f'{example.id}.inprogress'
        )
        pg.io.writefile(
            inprogress_file,
            datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        )
      runner.background_run(_save_inprogress_file, example)

  def on_example_complete(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    if example.newly_processed or runner.current_run.force_recompute_metrics:
      self._save_example(runner, experiment, example)

  def _load_experiment(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    """Creates the checkpoint file."""
    ckpt_files = self._list_checkpoint_files(runner, experiment)
    experiment.info(f'Found {len(ckpt_files)} checkpoint files to load.')

    # Load the checkpoint files in parallel.
    current_run = runner.current_run
    examples_to_load = current_run.examples_to_load(experiment)
    examples_to_load_metadata = current_run.examples_to_load_metadata(
        experiment
    )
    context = dict(counter=0, counter_lock=threading.Lock())

    def _load_state(ckpt_file):
      error = None
      with pg.timeit() as t:
        try:
          loaded_examples = experiment.load_state(
              ckpt_file,
              filter=lambda x: x.id in examples_to_load,
              load_example_metadata=lambda x: x.id in examples_to_load_metadata,
          )
        except BaseException as e:  # pylint: disable=broad-except
          loaded_examples = []
          error = e
        finally:
          with context['counter_lock']:
            context['counter'] += 1

          progress_str = f'{context["counter"]}/{len(ckpt_files)}'
          if error is None:
            experiment.info(
                f'Checkpoint file {ckpt_file!r} loaded in {t.elapse:.2f} '
                f'seconds. ({progress_str})'
            )
          else:
            experiment.warning(
                f'Failed to load checkpoint file {ckpt_file!r}: {error}. '
                f'Skipping the file. ({progress_str})'
            )

        output_ckpt_file = current_run.output_path_for(
            experiment, os.path.basename(ckpt_file)
        )
        if (not runner.current_run.force_recompute_metrics
            and ckpt_file != output_ckpt_file
            and any(e for e in loaded_examples if not e.has_error)):
          # Write the error-free warm-start examples to the output checkpoint
          # file.
          with SequenceWriter(output_ckpt_file) as writer:
            for example in loaded_examples:
              if not example.has_error:
                writer.add(example)

    _ = list(
        lf.concurrent_map(
            _load_state,
            ckpt_files,
            max_workers=self.max_ckpt_loading_threads,
            silence_on_errors=None
        )
    )

  @abc.abstractmethod
  def _list_checkpoint_files(
      self, runner: Runner, experiment: Experiment
  ) -> list[str]:
    """Lists the checkpoint file paths to restore."""

  @abc.abstractmethod
  def _save_example(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves an evaluated example."""


class PerExampleCheckpointer(Checkpointer):
  """Checkpointer that saves each example to a separate file.

  This checkpointer saves each processed example to its own checkpoint file,
  named using the pattern `<checkpoint_filename_prefix>_<example_id>.<ext>`.
  For example, `checkpoint_1.bagz`, `checkpoint_2.bagz`, etc.
  """

  def _on_bound(self):
    super()._on_bound()
    prefix, ext = self._file_prefix_and_ext(self.checkpoint_filename)
    self._checkpoint_file_prefix = prefix
    self._checkpoint_file_ext = ext

  def _list_checkpoint_files(
      self, runner: Runner, experiment: Experiment
  ) -> list[str]:

    def _list_checkpoints_from(ckpt_dir: str, examples_to_load: set[int]):
      ckpt_files = []
      if pg.io.path_exists(ckpt_dir):
        regex = re.compile(
            f'{self._checkpoint_file_prefix}_(\\d+){self._checkpoint_file_ext}'
            .replace('.', '\\.')
        )
        for filename in pg.io.listdir(ckpt_dir):
          match = regex.match(filename)
          if match and int(match.group(1)) in examples_to_load:
            examples_to_load.remove(int(match.group(1)))
            ckpt_files.append(os.path.join(ckpt_dir, filename))
      return ckpt_files

    examples_to_load = runner.current_run.examples_to_load(experiment)

    # Take output directory as the first priority to checkpoints processed in
    # this run.
    ckpt_files = _list_checkpoints_from(
        runner.current_run.output_dir(experiment), examples_to_load
    )
    # If the input and output directories are different, also load from the
    # input directory.
    if (examples_to_load
        and runner.current_run.input_root != runner.current_run.output_root):
      ckpt_files.extend(
          _list_checkpoints_from(
              runner.current_run.input_dir(experiment), examples_to_load
          )
      )
    return ckpt_files

  def _save_example(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    def save_state(example: Example):
      writer = SequenceWriter(
          runner.current_run.output_path_for(
              experiment,
              (
                  f'{self._checkpoint_file_prefix}_{example.id}'
                  f'{self._checkpoint_file_ext}'
              )
          )
      )
      try:
        writer.add(example)
        writer.close()
        experiment.info(
            f'Example {example.id} checkpointed to {writer.path}.',
        )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to checkpoint example {example.id} to {writer.path}. '
            f'Error: {e}, Stacktrace: \n{traceback.format_exc()}.',
        )
        raise e
    runner.background_run(save_state, example)

  def _file_prefix_and_ext(self, filename: str) -> tuple[str, str]:
    ext_index = filename.rfind('.')
    if ext_index == -1:
      return filename, ''
    else:
      return filename[:ext_index], filename[ext_index:]


class BulkCheckpointer(Checkpointer):
  """Checkpointer that saves all examples of an evaluation to a single file.

  This checkpointer appends newly processed examples of an evaluation to a
  single sequence file (e.g., `checkpoint.bagz`). This is often more efficient
  than `PerExampleCheckpointer` when dealing with a large number of examples
  or when file system overhead is a concern.
  """

  def _on_bound(self):
    super()._on_bound()
    self._lock = threading.Lock()
    self._sequence_writer = None

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment,
  ) -> None:
    self._sequence_writer = {}

  def on_run_abort(
      self,
      runner: Runner,
      root: Experiment,
      error: BaseException
  ) -> None:
    with self._lock:
      if self._sequence_writer is not None:
        for writer in self._sequence_writer.values():
          writer.close()
        self._sequence_writer.clear()

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment,
  ) -> None:
    with self._lock:
      assert self._sequence_writer is not None and not self._sequence_writer

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    super().on_experiment_start(runner, experiment)

    # Prepare the sequence writer for the experiment.
    if experiment.is_leaf:
      sequence_writer = SequenceWriter(
          runner.current_run.output_path_for(
              experiment, self.checkpoint_filename
          )
      )
      with self._lock:
        if self._sequence_writer is not None:
          self._sequence_writer[experiment.id] = sequence_writer

  def _list_checkpoint_files(
      self, runner: Runner, experiment: Experiment
  ) -> list[str]:
    # Always honor the output directory if it's present, as it contains both
    # the warm-started examples and newly processed examples.
    output_ckpt_file = runner.current_run.output_path_for(
        experiment, self.checkpoint_filename
    )
    if pg.io.path_exists(output_ckpt_file):
      return [output_ckpt_file]

    if runner.current_run.input_root != runner.current_run.output_root:
      input_ckpt_file = runner.current_run.input_path_for(
          experiment, self.checkpoint_filename
      )
      if pg.io.path_exists(input_ckpt_file):
        return [input_ckpt_file]
    return []

  def on_experiment_complete(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    """Closes the checkpoint file."""
    if not experiment.is_leaf:
      return
    assert experiment.id in self._sequence_writer
    with self._lock:
      if self._sequence_writer is not None:
        # Make sure the writer is closed without delay so the file will be
        # available immediately.
        writer = self._sequence_writer.pop(experiment.id)
        writer.close()
        experiment.info(
            f'{len(experiment.state.evaluation_status)} examples are '
            f'checkpointed to {writer.path}.'
        )

  def _save_example(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    assert experiment.id in self._sequence_writer
    def _save_example(example: Example):
      writer = self._sequence_writer[experiment.id]
      try:
        writer.add(example)
        experiment.info(
            f'Example {example.id} checkpointed to {writer.path}.',
        )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to checkpoint example {example.id} to {writer.path}. '
            f'Error: {e}, Stacktrace: \n{traceback.format_exc()}.',
        )
        raise e
    runner.background_run(_save_example, example)


class SequenceWriter:
  """A thread-safe writer for sequence files (e.g., Bagz) with atomic write.

  `SequenceWriter` wraps a `pg.io.SequenceWriter` to provide thread-safe
  `add` and `close` operations, ensuring that examples can be written
  concurrently from multiple threads without corrupting the sequence file.
  It writes to a temporary file and renames it to target path on `close` to
  achieve atomic write. If the target path exists, new examples are appended
  to existing content.
  """

  def __init__(self, path: str):
    self._lock = threading.Lock()
    self._path = path
    basename = os.path.basename(path)
    self._tmp_path = os.path.join(
        os.path.dirname(path), f'tmp.{basename}'
    )
    if pg.io.path_exists(self._path):
      pg.io.copy(self._path, self._tmp_path)
    self._sequence_writer = pg.io.open_sequence(self._tmp_path, 'a')

  @property
  def path(self) -> str:
    return self._path

  def add(self, example: Example):
    example_blob = pg.to_json_str(
        example,
        hide_default_values=True,
        save_ref_value=True,
        exclude_input=False,
    )
    with self._lock:
      if self._sequence_writer is None:
        return
      self._sequence_writer.add(example_blob)

  def close(self):
    # Make sure there is no write in progress.
    with self._lock:
      if self._sequence_writer is None:
        return
      self._sequence_writer.close()
      self._sequence_writer = None
      pg.io.rename(self._tmp_path, self._path)

  def __enter__(self):
    return self

  def __exit__(self, *args, **kwargs):
    del args, kwargs
    self.close()

  def __del__(self):
    self.close()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gemini API message conversion."""

import base64
from typing import Annotated, Any, Callable

import langfun.core as lf
from langfun.core import modalities as lf_modalities


class GeminiMessageConverter(lf.MessageConverter):
  """Converter between Langfun messages and Gemini API message format.

  This converter translates `lf.Message` objects into the JSON format required
  by the public Gemini API (e.g., via Vertex AI or Google AI Studio) and
  vice versa. It handles text and modalities like images, extracting thought
  chunks if present. An optional `chunk_preprocessor` can be provided to
  modify or filter chunks before conversion.
  """

  FORMAT_ID = 'gemini'

  chunk_preprocessor: Annotated[
      Callable[[str | lf.Modality], Any] | None,
      (
          'Chunk preprocessor for Langfun chunk to Gemini chunk conversion. '
          'It will be applied before each Langfun chunk is converted. '
          'If it returns None, the chunk will be skipped.'
      ),
  ] = None

  def to_value(
      self, message: lf.Message, always_send_content: bool = True
  ) -> dict[str, Any]:
    """Converts a Langfun message to Gemini API."""
    parts = []
    for chunk in message.chunk():
      if self.chunk_preprocessor:
        chunk = self.chunk_preprocessor(chunk)
        if chunk is None:
          continue

      if isinstance(chunk, str):
        parts.append(self._convert_chunk(chunk, always_send_content))
      else:
        if isinstance(chunk, lf_modalities.Mime):
          modalities = [chunk]
        # NOTE(daiyip): preprocessing may convert a single chunk into
        # a list of chunks
        elif isinstance(chunk, list):
          modalities = chunk
        else:
          raise ValueError(f'Unsupported content type: {chunk!r}.')
        parts.extend(
            self._convert_chunk(c, always_send_content) for c in modalities
        )
    return dict(role=self.get_role(message), parts=parts)

  def _convert_chunk(
      self, chunk: str | lf.Modality, always_send_content: bool = True
  ) -> Any:
    """Converts a Langfun chunk to Gemini chunk."""
    if isinstance(chunk, str):
      return {'text': chunk}
    if not isinstance(chunk, lf_modalities.Mime):
      raise ValueError(f'Unsupported content chunk: {chunk!r}.')
    # NOTE(daiyip): special handling for YouTube video.
    if chunk.uri and chunk.uri.startswith('https://www.youtube.com/watch?v='):
      return {
          'fileData': {
              'mimeType': 'video/*',
              'fileUri': chunk.uri
          }
      }
    if chunk.is_text:
      return {'text': chunk.to_text()}
    if (
        not always_send_content
        and chunk.uri
        and chunk.uri.lower().startswith(('http:', 'https:', 'ftp:'))
    ):
      return {
          'fileData': {
              'mimeType': chunk.mime_type,
              'fileUri': chunk.uri,
          }
      }
    return {
        'inlineData': {
            'data': base64.b64encode(chunk.to_bytes()).decode(),
            'mimeType': chunk.mime_type,
        }
    }

  def from_value(self, value: dict[str, Any]) -> lf.Message:
    """Returns a Langfun message from Gemini message."""
    message_cls = self.get_message_cls(
        self._safe_read(value, 'role', default='model')
    )
    parts = self._safe_read(value, 'parts', default=[])
    assert isinstance(parts, list)

    chunks = []
    thought_chunks = []
    for part in parts:
      if 'text' in part:
        text = self._safe_read(part, 'text')
        if 'thought' in part:
          thought_chunks.append(text)
        else:
          chunks.append(text)
      elif 'thought' in part:
        # We occasionally encounter 'thought' without text.
        pass
      elif 'inlineData' in part:
        data = self._safe_read(part, 'inlineData')
        chunks.append(
            lf_modalities.Mime.class_from_mime_type(
                self._safe_read(data, 'mimeType')
            ).from_bytes(base64.b64decode(self._safe_read(data, 'data')))
        )
      elif 'fileData' in part:
        data = self._safe_read(part, 'fileData')
        chunks.append(
            lf_modalities.Mime.class_from_mime_type(
                self._safe_read(data, 'mimeType')
            ).from_uri(self._safe_read(data, 'fileUri'))
        )
      elif 'functionCall' in part or 'functionResponse' in part:
        pass
      else:
        raise ValueError(f'Unsupported content part: {part!r}.')
    message = message_cls.from_chunks(chunks)
    if thought_chunks:
      message.set('thought', message_cls.from_chunks(thought_chunks))
    return message

  @classmethod
  def get_role(cls, message: lf.Message) -> str:
    """Returns the role of the message."""
    if isinstance(message, lf.AIMessage):
      return 'model'
    return super().get_role(message)

  @classmethod
  def get_message_cls(cls, role: str) -> type[lf.Message]:
    """Returns the message class of the message."""
    if role == 'model':
      return lf.AIMessage
    return super().get_message_cls(role)


def _as_gemini_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns a Gemini (REST) format message."""
  return GeminiMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_gemini_format(
    cls,
    gemini_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the Gemini (REST) format message."""
  del cls
  return GeminiMessageConverter(**kwargs).from_value(gemini_message)

# Set shortcut methods in lf.Message.
lf.Message.as_gemini_format = _as_gemini_format
lf.Message.from_gemini_format = _from_gemini_format



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Schema and Prompting Protocol for Structured Data."""

import abc
import inspect
import io
import textwrap
import typing
from typing import Any, ClassVar, Type, Union
import langfun.core as lf
import pyglove as pg


def _parse_value_spec(value) -> pg.typing.ValueSpec:
  """Parses a PyGlove ValueSpec equivalent into a ValueSpec.

  Examples:
  ```
  _parse_value_spec(int) -> pg.typing.Int
  _parse_value_spec(list[int]) -> pg.typing.List(pg.typing.Int)
  _parse_value_spec(dict(a=int, b=str)) -> pg.typing.Dict(
      pg.typing.Int, pg.typing.Str
  )
  ```
  Args:
    value: The value to parse. It can be a PyGlove ValueSpec, a dict with a
      single 'result' key, or a Python type annotation.

  Returns:
    A PyGlove ValueSpec.
  """
  if isinstance(value, pg.typing.ValueSpec):
    return value

  if isinstance(value, dict) and len(value) == 1 and 'result' in value:
    value = value['result']

  def _parse_node(v) -> pg.typing.ValueSpec:
    if isinstance(v, dict):
      return pg.typing.Dict([(k, _parse_node(cv)) for k, cv in v.items()])
    elif isinstance(v, list):
      if len(v) != 1:
        raise ValueError(
            'Annotation with list must be a list of a single element. '
            f'Encountered: {v}'
        )
      return pg.typing.List(_parse_node(v[0]))
    else:
      spec = pg.typing.ValueSpec.from_annotation(v, auto_typing=True)
      if isinstance(
          spec,
          (
              pg.typing.Any,
              pg.typing.Callable,
              pg.typing.Tuple,
              pg.typing.Type,
              pg.typing.Union,
          ),
      ):
        raise ValueError(f'Unsupported schema specification: {v}')
      return spec

  return _parse_node(value)


class SchemaError(Exception):   # pylint: disable=g-bad-exception-name
  """Schema error."""

  def __init__(
      self,
      schema: 'Schema',
      value: Any,
      protocol: str,
      cause: Exception
  ):
    self.schema = schema
    self.value = value
    self.protocol = protocol
    self.cause = cause

  def __str__(self):
    r = io.StringIO()
    r.write(
        pg.colored(
            f'{self.cause.__class__.__name__}: {self.cause}', 'magenta'
        )
    )

    r.write('\n')
    r.write(pg.colored('Schema:', 'red'))
    r.write('\n\n')
    r.write(textwrap.indent(
        pg.colored(
            schema_repr(self.schema, protocol=self.protocol), 'magenta'
        ),
        ' ' * 2
    ))
    r.write('\n\n')
    r.write(pg.colored('Generated value:', 'red'))
    r.write('\n\n')
    r.write(textwrap.indent(
        pg.colored(value_repr(self.value, protocol=self.protocol), 'magenta'),
        ' ' * 2
    ))
    return r.getvalue()


class Schema(
    lf.NaturalLanguageFormattable,
    pg.Object,
    pg.views.HtmlTreeView.Extension
):
  """Schema for structured inputs and outputs.

  `lf.Schema` provides a unified representation for defining the output schema
  used in Langfun's structured operations like `lf.query`, `lf.parse`,
  `lf.complete`, and `lf.describe`. It acts as an abstraction layer,
  allowing schemas to be defined using Python type annotations, `pg.Object`
  classes, or dictionaries, and then converting them into a format that
  language models can understand.

  `lf.Schema` can be created from various types using `lf.Schema.from_value`:
  *   Built-in types: `int`, `str`, `bool`, `float`
  *   Typing constructs: `list`, `dict`, `typing.Union`, `typing.Literal`,
      `typing.Optional`
  *   PyGlove classes: `pg.Object` subclasses

  **1. Creating a Schema:**

  ```python
  import langfun as lf
  import pyglove as pg
  from typing import Literal, Union

  # From a basic type
  int_schema = lf.Schema.from_value(int)

  # From a list type
  list_schema = lf.Schema.from_value(list[int])

  # From a dictionary
  dict_schema = lf.Schema.from_value(dict(a=int, b=str))

  # From pg.Object
  class Point(pg.Object):
    x: int
    y: int
  point_schema = lf.Schema.from_value(Point)

  # From Union or Literal
  union_schema = lf.Schema.from_value(Union[int, str])
  literal_schema = lf.Schema.from_value(Literal['A', 'B'])
  ```

  **2. Schema Representation:**
  Once created, a schema object can represent itself in different formats,
  such as Python-like syntax or JSON, which is used in prompts to LLMs.

  ```python
  print(point_schema.repr('python'))
  # Output:
  # class Point:
  #   x: int
  #   y: int

  print(dict_schema.repr('json'))
  # Output:
  # {
  #   "a": "int",
  #   "b": "str"
  # }
  ```
  """

  spec: pg.typing.Annotated[
      pg.typing.Object(pg.typing.ValueSpec, transform=_parse_value_spec),
      (
          'A PyGlove ValueSpec object representing the spec for the value '
          'to be parsed.'
      ),
  ]

  def schema_repr(self, protocol: str = 'python', **kwargs) -> str:
    """Returns the representation of the schema."""
    return schema_repr(self, protocol=protocol, **kwargs)

  def value_repr(
      self, value: Any, protocol: str = 'python', **kwargs
  ) -> str:
    """Returns the representation of a structured value."""
    return value_repr(value, schema=self, protocol=protocol, **kwargs)

  def parse_value(
      self, text: str, protocol: str = 'python', **kwargs
  ) -> Any:
    """Parses a LM generated text into a structured value."""
    value = parse_value(text, schema=self, protocol=protocol, **kwargs)

    # TODO(daiyip): support autofix for schema error.
    try:
      return self.spec.apply(value)
    except Exception as e:
      raise SchemaError(self, value, protocol, e)  # pylint: disable=raise-missing-from

  def natural_language_format(self) -> str:
    return self.schema_str()

  def schema_dict(self) -> dict[str, Any]:
    """Returns the dictionary representation of the schema."""

    def _node(vs: pg.typing.ValueSpec) -> Any:
      if isinstance(vs, pg.typing.PrimitiveType):
        return vs
      elif isinstance(vs, pg.typing.Dict):
        assert vs.schema is not None
        return {str(k): _node(f.value) for k, f in vs.schema.fields.items()}
      elif isinstance(vs, pg.typing.List):
        return [_node(vs.element.value)]
      elif isinstance(vs, pg.typing.Object):
        if issubclass(vs.cls, pg.Object):
          d = {pg.JSONConvertible.TYPE_NAME_KEY: vs.cls.__serialization_key__}
          d.update(
              {
                  str(k): _node(f.value)
                  for k, f in vs.cls.__schema__.fields.items()
              }
          )
          return d
      raise TypeError(
          'Unsupported value spec being used as the schema for '
          f'structured data: {vs}.')

    return {'result': _node(self.spec)}

  def class_dependencies(
      self,
      include_base_classes: bool = True,
      include_subclasses: bool = True,
      include_generated_subclasses: bool = False) -> list[Type[Any]]:
    """Returns a list of class dependencies for current schema."""
    return class_dependencies(
        self.spec,
        include_base_classes,
        include_subclasses,
        include_generated_subclasses
    )

  @classmethod
  def from_value(cls, value) -> 'Schema':
    """Creates a schema from an equivalent representation."""
    if isinstance(value, Schema):
      return value
    return cls(_parse_value_spec(value))

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      **kwargs,
  ):
    return pg.Html.element(
        'div',
        [pg.Html.escape(self.schema_repr(protocol='python'))],
        css_classes=['lf-schema-definition']
    ).add_style(
        """
        .lf-schema-definition {
            color: blue;
            margin: 5px;
            white-space: pre-wrap;
        }
        """
    )


SchemaType = Union[Schema, Type[Any], list[Type[Any]], dict[str, Any]]


def _top_level_object_specs_from_value(value: pg.Symbolic) -> list[Type[Any]]:
  """Returns a list of top level value specs from a symbolic value."""
  top_level_object_specs = []

  def _collect_top_level_object_specs(k, v, p):
    del k, p
    if isinstance(v, pg.Object):
      top_level_object_specs.append(pg.typing.Object(v.__class__))
      return pg.TraverseAction.CONTINUE
    return pg.TraverseAction.ENTER

  pg.traverse(value, _collect_top_level_object_specs)
  return top_level_object_specs


def class_dependencies(
    value_or_spec: Union[
        pg.Symbolic,
        Schema,
        pg.typing.ValueSpec,
        Type[pg.Object],
        tuple[Union[pg.typing.ValueSpec, Type[pg.Object]], ...],
    ],
    include_base_classes: bool = True,
    include_subclasses: bool = True,
    include_generated_subclasses: bool = False,
) -> list[Type[Any]]:
  """Returns a list of class dependencies from a value or specs."""
  if isinstance(value_or_spec, Schema):
    value_or_spec = value_or_spec.spec

  if inspect.isclass(value_or_spec) or isinstance(
      value_or_spec, pg.typing.ValueSpec
  ):
    value_or_spec = (value_or_spec,)

  if isinstance(value_or_spec, tuple):
    value_specs = []
    for v in value_or_spec:
      if isinstance(v, pg.typing.ValueSpec):
        value_specs.append(v)
      elif inspect.isclass(v):
        value_specs.append(pg.typing.Object(v))
      else:
        raise TypeError(f'Unsupported spec type: {v!r}')
  else:
    value_specs = _top_level_object_specs_from_value(value_or_spec)

  seen = set()
  dependencies = []

  def _add_dependency(cls_or_classes):
    if isinstance(cls_or_classes, type):
      cls_or_classes = [cls_or_classes]
    for cls in cls_or_classes:
      if cls not in dependencies:
        dependencies.append(cls)

  def _fill_dependencies(vs: pg.typing.ValueSpec, include_subclasses: bool):
    if isinstance(vs, pg.typing.Object):
      cls = vs.cls
      if cls.__module__ == 'builtins':
        return

      if cls not in seen:
        seen.add(cls)

        if include_base_classes:
          # Add base classes as dependencies.
          for base_cls in cls.__bases__:
            # We only keep track of user-defined symbolic classes.
            if base_cls is not object and base_cls is not pg.Object:
              _fill_dependencies(
                  pg.typing.Object(base_cls), include_subclasses=False
              )

        # Add members as dependencies.
        for field in pg.schema(cls).values():
          _fill_dependencies(field.value, include_subclasses)
      _add_dependency(cls)

      # Check subclasses if available.
      if include_subclasses:
        for subcls in cls.__subclasses__():
          # NOTE(daiyip): To prevent LLM-generated "hallucinated" classes from
          # polluting the generation space, classes dynamically created by
          # 'eval' (which have __module__ == 'builtins') are excluded from
          # dependencies by default.
          if ((include_generated_subclasses or subcls.__module__ != 'builtins')
              and subcls not in dependencies):
            _fill_dependencies(
                pg.typing.Object(subcls), include_subclasses=True
            )

    if isinstance(vs, pg.typing.List):
      _fill_dependencies(vs.element.value, include_subclasses)
    elif isinstance(vs, pg.typing.Tuple):
      for elem in vs.elements:
        _fill_dependencies(elem.value, include_subclasses)
    elif isinstance(vs, pg.typing.Dict) and vs.schema:
      for v in vs.schema.values():
        _fill_dependencies(v.value, include_subclasses)
    elif isinstance(vs, pg.typing.Union):
      for v in vs.candidates:
        _fill_dependencies(v, include_subclasses)

  for value_spec in value_specs:
    _fill_dependencies(value_spec, include_subclasses)
  return dependencies


def schema_spec(noneable: bool = False) -> pg.typing.ValueSpec:  # pylint: disable=unused-argument
  if typing.TYPE_CHECKING:
    return Any
  return pg.typing.Object(
      Schema, transform=Schema.from_value, is_noneable=noneable
  )  # pylint: disable=unreachable-code


def annotation(
    vs: pg.typing.ValueSpec,
    annotate_optional: bool = True,
    strict: bool = False,
    allowed_dependencies: set[Type[Any]] | None = None,
) -> str:
  """Returns the annotation string for a value spec."""
  child_annotation_kwargs = dict(
      strict=strict, allowed_dependencies=allowed_dependencies
  )
  if isinstance(vs, pg.typing.Any):
    return 'Any'
  elif isinstance(vs, pg.typing.Enum):
    candidate_str = ', '.join([repr(v) for v in vs.values])
    return f'Literal[{candidate_str}]'
  elif isinstance(vs, pg.typing.Union):
    candidate_str = ', '.join(
        [
            annotation(c, annotate_optional=False, **child_annotation_kwargs)
            for c in vs.candidates
        ]
    )
    if vs.is_noneable:
      candidate_str += ', None'
    return f'Union[{candidate_str}]'

  if isinstance(vs, pg.typing.Bool):
    x = 'bool'
  elif isinstance(vs, pg.typing.Str):
    if vs.regex is None:
      x = 'str'
    else:
      if strict:
        x = f"pg.typing.Str(regex='{vs.regex.pattern}')"
      else:
        x = f"str(regex='{vs.regex.pattern}')"
  elif isinstance(vs, pg.typing.Number):
    constraints = []
    min_label = 'min_value' if strict else 'min'
    max_label = 'max_value' if strict else 'max'
    if vs.min_value is not None:
      constraints.append(f'{min_label}={vs.min_value}')
    if vs.max_value is not None:
      constraints.append(f'{max_label}={vs.max_value}')
    x = 'int' if isinstance(vs, pg.typing.Int) else 'float'
    if constraints:
      if strict:
        x = (
            'pg.typing.Int'
            if isinstance(vs, pg.typing.Int)
            else 'pg.typing.Float'
        )
      x += '(' + ', '.join(constraints) + ')'
  elif isinstance(vs, pg.typing.Object):
    if allowed_dependencies is None or vs.cls in allowed_dependencies:
      x = vs.cls.__name__
    else:
      x = 'Any'
  elif isinstance(vs, pg.typing.List):
    item_str = annotation(vs.element.value, **child_annotation_kwargs)
    x = f'list[{item_str}]'
  elif isinstance(vs, pg.typing.Tuple):
    elem_str = ', '.join(
        [annotation(el.value, **child_annotation_kwargs) for el in vs.elements]
    )
    x = f'tuple[{elem_str}]'
  elif isinstance(vs, pg.typing.Dict):
    kv_pairs = None
    if vs.schema is not None:
      kv_pairs = [
          (k, annotation(f.value, **child_annotation_kwargs))
          for k, f in vs.schema.items()
          if isinstance(k, pg.typing.ConstStrKey)
      ]

    if kv_pairs:
      kv_str = ', '.join(f"'{k}': {v}" for k, v in kv_pairs)
      x = '{' + kv_str + '}'
      if strict:
        x = f'pg.typing.Dict({x})'
    elif vs.schema and vs.schema.dynamic_field:
      v = annotation(vs.schema.dynamic_field.value, **child_annotation_kwargs)
      x = f'dict[str, {v}]'
    else:
      x = 'dict[str, Any]'

  else:
    raise TypeError(f'Unsupported value spec being used as schema: {vs}.')

  if annotate_optional and vs.is_noneable:
    x += ' | None'
  return x

#
# Prompting protocols for structured data.
#


class PromptingProtocol(metaclass=abc.ABCMeta):
  """Base class for prompting protocols for structured data."""

  NAME: ClassVar[str]

  _PROTOCOLS: ClassVar[dict[str, Type['PromptingProtocol']]] = {}

  def __init_subclass__(cls):
    PromptingProtocol._PROTOCOLS[cls.NAME] = cls

  @classmethod
  def from_name(cls, name: str) -> 'PromptingProtocol':
    """Returns the prompting protocol from the name."""
    protocol_cls = cls._PROTOCOLS.get(name)
    if protocol_cls is None:
      raise ValueError(f'Unsupported protocol: {name}.')
    return protocol_cls()  # pytype: disable=not-instantiable

  @abc.abstractmethod
  def schema_repr(self, schema: Schema) -> str:
    """Returns the representation of the schema."""

  @abc.abstractmethod
  def value_repr(
      self,
      value: Any,
      schema: Schema | None = None,
      **kwargs
  ) -> str:
    """Returns the representation of a structured value."""

  @abc.abstractmethod
  def parse_value(
      self,
      text: str,
      schema: Schema | None = None,
      **kwargs
  ) -> Any:
    """Parses a LM generated text into a structured value."""


def schema_repr(
    schema: Schema,
    *,
    protocol: str = 'python',
    **kwargs
) -> str:
  """Returns the representation of the schema based on the protocol."""
  return PromptingProtocol.from_name(protocol).schema_repr(schema, **kwargs)


def value_repr(
    value: Any,
    schema: Schema | None = None,
    *,
    protocol: str = 'python',
    **kwargs) -> str:
  """Returns the representation of a structured value based on the protocol."""
  return PromptingProtocol.from_name(protocol).value_repr(
      value, schema, **kwargs
  )


def parse_value(
    text: str,
    schema: Schema | None = None,
    *,
    protocol: str = 'python',
    **kwargs
) -> Any:
  """Parses a LM generated text into a structured value."""
  return PromptingProtocol.from_name(protocol).parse_value(
      text, schema=schema, **kwargs
  )


#
# Special value markers.
#


class Missing(pg.Object, pg.typing.CustomTyping):
  """Value marker for a missing field.

  This class differs from pg.MISSING_VALUE in two aspects:
  * When a field is assigned with lf.Missing(), it's considered non-partial.
  * lf.Missing() could format the value spec as Python annotations that are
    consistent with `lf.structured.Schema.schema_repr()`.
  """

  def _on_bound(self):
    super()._on_bound()
    self._value_spec = None

  @property
  def value_spec(self) -> pg.ValueSpec | None:
    """Returns the value spec that applies to the current missing value."""
    return self._value_spec

  def custom_apply(
      self, path: pg.KeyPath, value_spec: pg.ValueSpec, *args, **kwargs
  ) -> tuple[bool, Any]:
    self._value_spec = value_spec
    return (False, self)

  def format(self, *args, **kwargs) -> str:
    if self._value_spec is None:
      return 'MISSING'
    return f'MISSING({annotation(self._value_spec)})'

  @classmethod
  def find_missing(cls, value: Any) -> dict[str, 'Missing']:
    """Lists all missing values contained in the value."""
    missing = {}

    def _visit(k, v, p):
      del p
      if isinstance(v, Missing):
        missing[k] = v
      return pg.TraverseAction.ENTER

    pg.traverse(value, _visit)
    return missing


MISSING = Missing()


def mark_missing(value: Any) -> Any:
  """Replaces pg.MISSING within the value with lf.structured.Missing objects."""
  if isinstance(value, list):
    value = pg.List(value)
  elif isinstance(value, dict):
    value = pg.Dict(value)
  if isinstance(value, pg.Symbolic):

    def _mark_missing(k, v, p):
      del k, p
      if pg.MISSING_VALUE == v:
        v = Missing()
      return v

    return value.rebind(_mark_missing, raise_on_no_change=False)
  return value


class Unknown(pg.Object, pg.typing.CustomTyping):
  """Value marker for a field that LMs could not provide."""

  def custom_apply(self, *args, **kwargs) -> tuple[bool, Any]:
    return (False, self)

  def format(self, *args, **kwargs) -> str:
    return 'UNKNOWN'


UNKNOWN = Unknown()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In-memory LM cache."""

import collections
import contextlib
import json
from typing import Annotated, Any, Iterator
import langfun.core as lf
from langfun.core.llms.cache import base
import pyglove as pg


@pg.use_init_args(['filename', 'ttl', 'key'])
class InMemory(base.LMCacheBase):
  """An in-memory cache for language model lookups.

  `InMemory` stores LM prompts and their corresponding responses in memory,
  providing a simple and fast caching mechanism for a single session.
  Optionally, it can persist the cache to a JSON file on disk, allowing
  results to be reused across sessions.

  When a filename is provided, the cache will be loaded from the file upon
  initialization and saved to the file when `save()` is called. This is
  useful for caching results in interactive environments like Colab or
  when running batch jobs.

  Example:

  ```python
  import langfun as lf
  # Using in-memory cache without persistence
  lm = lf.llms.GeminiPro(cache=lf.llms.cache.InMemory())
  r = lm.query('hello')

  # Using in-memory cache with persistence
  lm = lf.llms.GeminiPro(cache=lf.llms.cache.InMemory('cache.json'))
  r = lm.query('hello')
  lm.cache.save()
  ```
  """

  filename: Annotated[
      str | None,
      (
          'File name to load and save in memory cache.'
      )
  ] = None

  def _on_bound(self) -> None:
    super()._on_bound()
    self._cache = collections.defaultdict(dict)

    if self.filename is not None:
      try:
        records = pg.load(self.filename)
        for record in records:
          model_cache = {}
          for entry in record.entries:
            model_cache[entry.k] = entry.v
          self._cache[record.model_id] = model_cache
      except FileNotFoundError:
        pg.logging.warning(
            "Creating a new cache as cache file '%s' does not exist.",
            self.filename,
        )
      except json.JSONDecodeError:
        pg.logging.warning(
            "Creating a new cache as cache file '%s' is corrupted.",
            self.filename,
        )

  def model_ids(self) -> list[str]:
    """Returns the model ids of cached queires."""
    return list(self._cache.keys())

  def __len__(self) -> int:
    """Returns the number of entries in the cache."""
    return sum(len(v) for v in self._cache.values())

  def keys(self, model_id: str | None = None) -> Iterator[str]:
    """Returns the cached keys for a model."""
    if model_id is None:
      for model_cache in self._cache.values():
        for k in model_cache.keys():
          yield k
    else:
      for k in self._cache[model_id].keys():
        yield k

  def values(self, model_id: str | None = None) -> Iterator[base.LMCacheEntry]:
    """Returns the cached entries for a model."""
    if model_id is None:
      for model_cache in self._cache.values():
        for v in model_cache.values():
          yield v
    else:
      for v in self._cache[model_id].values():
        yield v

  def items(
      self,
      model_id: str | None = None
      ) -> Iterator[tuple[str, base.LMCacheEntry]]:
    """Returns the cached items for a model."""
    if model_id is None:
      for model_cache in self._cache.values():
        for k, v in model_cache.items():
          yield k, v
    else:
      for k, v in self._cache[model_id].items():
        yield k, v

  def _get(self, model_id: str, key: Any) -> base.LMCacheEntry | None:
    """Returns a LM cache entry associated with the key."""
    return self._cache[model_id].get(key, None)

  def _put(self, model_id: str, key: Any, entry: base.LMCacheEntry) -> None:
    """Puts a LM cache entry associated with the key."""
    self._cache[model_id][key] = entry

  def _delete(self, model_id: str, key: str) -> bool:
    """Deletes a LM cache entry associated with the key."""
    model_cache = self._cache.get(model_id, None)
    if model_cache is None:
      return False
    return model_cache.pop(key, None) is not None

  def reset(self, model_id: str | None = None) -> None:
    """Resets the cache."""
    if model_id is not None:
      self._cache[model_id].clear()
    else:
      self._cache.clear()

  def _sym_clone(self, deep: bool, memo: Any = None) -> 'InMemory':
    v = super()._sym_clone(deep, memo)
    v._cache = self._cache  # pylint: disable=protected-access
    return v

  def save(self, path: str | None = None) -> None:
    """Saves the in-memory cache."""
    if path is None:
      if self.filename is None:
        raise ValueError('`path` must be specified.')
      path = self.filename

    # Do nothing if there is no update, this avoids unnecessary rewrites.
    if self.stats.num_updates == 0 and path == self.filename:
      return

    records = []
    for model_id in self.model_ids():
      entries = [dict(k=k, v=v) for k, v in self.items(model_id)]
      records.append(dict(model_id=model_id, entries=entries))
    pg.save(records, path)


@contextlib.contextmanager
def lm_cache(filename: str | None = None) -> Iterator[InMemory]:
  """Context manager to enable in-memory cache for LMs in the current context.

  This context manager sets an `InMemory` cache as the default cache for
  any Langfun language model instantiated within its scope, unless a model
  is explicitly configured with a different cache.

  If a `filename` is provided, the cache will be loaded from the specified
  file at the beginning of the context and automatically saved back to the
  file upon exiting the context. This is a convenient way to manage
  persistent caching for a block of code.

  Example:

  ```python
  import langfun as lf
  with lf.lm_cache('my_cache.json'):
    # LMs created here will use 'my_cache.json' for caching.
    lm = lf.llms.GeminiPro()
    print(lm.query('hello'))
  ```

  Args:
    filename: If provided, specifies the JSON file for loading and saving
      the cache.

  Yields:
    The `InMemory` cache instance created for this context.
  """
  cache = InMemory(filename)
  try:
    with lf.context(cache=cache):
      yield cache
  finally:
    if filename is not None:
      cache.save()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun LLM implementations."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top

# LMs for testing.
from langfun.core.llms.fake import Fake
from langfun.core.llms.fake import Echo
from langfun.core.llms.fake import Pseudo
from langfun.core.llms.fake import StaticMapping
from langfun.core.llms.fake import StaticResponse
from langfun.core.llms.fake import StaticSequence

# Compositional models.
from langfun.core.llms.compositional import RandomChoice

# Base models by request/response protocol.
from langfun.core.llms.rest import REST
from langfun.core.llms.openai_compatible import OpenAIChatCompletionAPI
from langfun.core.llms.openai_compatible import OpenAIResponsesAPI
from langfun.core.llms.gemini import Gemini
from langfun.core.llms.anthropic import Anthropic

# Base models by serving platforms.
from langfun.core.llms.vertexai import VertexAI
from langfun.core.llms.groq import Groq
from langfun.core.llms.azure_openai import AzureOpenAI

# Gemini models.
from langfun.core.llms.google_genai import GenAI
from langfun.core.llms.google_genai import Gemini3ProPreview
from langfun.core.llms.google_genai import Gemini3FlashPreview
from langfun.core.llms.google_genai import Gemini25Pro
from langfun.core.llms.google_genai import Gemini25Flash
from langfun.core.llms.google_genai import Gemini25ProPreview_20250605
from langfun.core.llms.google_genai import Gemini25FlashPreview_20250520
from langfun.core.llms.google_genai import Gemini25ProPreview_20250506
from langfun.core.llms.google_genai import Gemini25FlashPreview_20250417
from langfun.core.llms.google_genai import Gemini25ProPreview_20250325
from langfun.core.llms.google_genai import Gemini2Flash
from langfun.core.llms.google_genai import Gemini2Flash_001
from langfun.core.llms.google_genai import Gemini2FlashLitePreview_20250205
from langfun.core.llms.google_genai import Gemini15Pro
from langfun.core.llms.google_genai import Gemini15Pro_002
from langfun.core.llms.google_genai import Gemini15Pro_001
from langfun.core.llms.google_genai import Gemini15Flash
from langfun.core.llms.google_genai import Gemini15Flash_002
from langfun.core.llms.google_genai import Gemini15Flash_001
from langfun.core.llms.google_genai import Gemini15Flash8B
from langfun.core.llms.google_genai import Gemini15Flash8B_001
from langfun.core.llms.google_genai import Gemini2ProExp_20250205
from langfun.core.llms.google_genai import Gemini2FlashThinkingExp_20250121
from langfun.core.llms.google_genai import GeminiExp_20241206
from langfun.core.llms.google_genai import Gemini25FlashImagePreview
from langfun.core.llms.google_genai import Gemini3ProImagePreview

from langfun.core.llms.vertexai import VertexAIGemini
from langfun.core.llms.vertexai import VertexAIGemini2Flash
from langfun.core.llms.vertexai import VertexAIGemini2Flash_001
from langfun.core.llms.vertexai import VertexAIGemini2FlashLitePreview_20250205
from langfun.core.llms.vertexai import VertexAIGemini15Pro
from langfun.core.llms.vertexai import VertexAIGemini15Pro_002
from langfun.core.llms.vertexai import VertexAIGemini15Pro_001
from langfun.core.llms.vertexai import VertexAIGemini15Flash
from langfun.core.llms.vertexai import VertexAIGemini15Flash_002
from langfun.core.llms.vertexai import VertexAIGemini15Flash_001
from langfun.core.llms.vertexai import VertexAIGemini15Flash8B
from langfun.core.llms.vertexai import VertexAIGemini15Flash8B_001

from langfun.core.llms.vertexai import VertexAIGemini2ProExp_20250205
from langfun.core.llms.vertexai import VertexAIGemini2FlashThinkingExp_20250121
from langfun.core.llms.vertexai import VertexAIGeminiExp_20241206
from langfun.core.llms.vertexai import VertexAIGemini25ProExp_20250325
from langfun.core.llms.vertexai import VertexAIGemini25ProPreview_20250325
from langfun.core.llms.vertexai import VertexAIGemini25FlashPreview_20250417
from langfun.core.llms.vertexai import VertexAIGemini25ProPreview_20250506
from langfun.core.llms.vertexai import VertexAIGemini25FlashPreview_20250520
from langfun.core.llms.vertexai import VertexAIGemini25ProPreview_20250605
from langfun.core.llms.vertexai import VertexAIGemini25Pro
from langfun.core.llms.vertexai import VertexAIGemini25Flash
from langfun.core.llms.vertexai import VertexAIGemini25FlashImagePreview
from langfun.core.llms.vertexai import VertexAIGemini3ProPreview
from langfun.core.llms.vertexai import VertexAIGemini3ProImagePreview
from langfun.core.llms.vertexai import VertexAIGemini3FlashPreview

# For backward compatibility.
GeminiPro1_5 = Gemini15Pro
GeminiFlash1_5 = Gemini15Flash
VertexAIGeminiPro1_5 = VertexAIGemini15Pro
VertexAIGeminiFlash1_5 = VertexAIGemini15Flash

# OpenAI models.
from langfun.core.llms.openai import OpenAI

from langfun.core.llms.openai import Gpt51
from langfun.core.llms.openai import Gpt5
from langfun.core.llms.openai import Gpt5Mini
from langfun.core.llms.openai import Gpt41
from langfun.core.llms.openai import GptO3
from langfun.core.llms.openai import GptO4Mini
from langfun.core.llms.openai import Gpt45Preview_20250227
from langfun.core.llms.openai import GptO3Mini
from langfun.core.llms.openai import GptO1
from langfun.core.llms.openai import GptO1Preview
from langfun.core.llms.openai import GptO1Preview_20240912
from langfun.core.llms.openai import GptO1Mini
from langfun.core.llms.openai import GptO1Mini_20240912

from langfun.core.llms.openai import Gpt4oMini
from langfun.core.llms.openai import Gpt4oMini_20240718
from langfun.core.llms.openai import Gpt4o
from langfun.core.llms.openai import Gpt4o_20241120
from langfun.core.llms.openai import Gpt4o_20240806
from langfun.core.llms.openai import Gpt4o_20240513

from langfun.core.llms.openai import Gpt4Turbo
from langfun.core.llms.openai import Gpt4Turbo_20240409
from langfun.core.llms.openai import Gpt4TurboPreview
from langfun.core.llms.openai import Gpt4TurboPreview_20240125
from langfun.core.llms.openai import Gpt4TurboPreview_20231106
from langfun.core.llms.openai import Gpt4
from langfun.core.llms.openai import Gpt4_20230613

from langfun.core.llms.openai import Gpt4_32K
from langfun.core.llms.openai import Gpt4_32K_20230613

from langfun.core.llms.openai import Gpt35Turbo
from langfun.core.llms.openai import Gpt35Turbo_20240125
from langfun.core.llms.openai import Gpt35Turbo_20231106
from langfun.core.llms.openai import Gpt35Turbo_20230613
from langfun.core.llms.openai import Gpt35Turbo16K
from langfun.core.llms.openai import Gpt35Turbo16K_20230613

# For backward compatibility.
Gpt4TurboPreview_0125 = Gpt4TurboPreview_20240125
Gpt4TurboPreview_1106 = Gpt4TurboPreview_20231106
Gpt4_0613 = Gpt4_20230613
Gpt4_32K_0613 = Gpt4_32K_20230613
Gpt35Turbo_0125 = Gpt35Turbo_20240125
Gpt35Turbo_1106 = Gpt35Turbo_20231106
Gpt35Turbo_0613 = Gpt35Turbo_20230613
Gpt35Turbo16K_0613 = Gpt35Turbo16K_20230613

from langfun.core.llms.openai import Gpt35

# Anthropic models.

from langfun.core.llms.anthropic import Claude45
from langfun.core.llms.anthropic import Claude45Haiku_20251001
from langfun.core.llms.anthropic import Claude45Sonnet_20250929
from langfun.core.llms.anthropic import Claude45Opus_20251101
from langfun.core.llms.anthropic import Claude4
from langfun.core.llms.anthropic import Claude4Sonnet_20250514
from langfun.core.llms.anthropic import Claude4Opus_20250514
from langfun.core.llms.anthropic import Claude37
from langfun.core.llms.anthropic import Claude37Sonnet_20250219
from langfun.core.llms.anthropic import Claude35Sonnet
from langfun.core.llms.anthropic import Claude35Sonnet_20241022
from langfun.core.llms.anthropic import Claude35Haiku
from langfun.core.llms.anthropic import Claude35Haiku_20241022
from langfun.core.llms.anthropic import Claude3Opus
from langfun.core.llms.anthropic import Claude3Opus_20240229
from langfun.core.llms.anthropic import Claude3Sonnet
from langfun.core.llms.anthropic import Claude3Sonnet_20240229
from langfun.core.llms.anthropic import Claude3Haiku
from langfun.core.llms.anthropic import Claude3Haiku_20240307

from langfun.core.llms.vertexai import VertexAIAnthropic
from langfun.core.llms.vertexai import VertexAIClaude45Haiku_20251001
from langfun.core.llms.vertexai import VertexAIClaude45Sonnet_20250929
from langfun.core.llms.vertexai import VertexAIClaude45Opus_20251101
from langfun.core.llms.vertexai import VertexAIClaude4Opus_20250514
from langfun.core.llms.vertexai import VertexAIClaude4Sonnet_20250514
from langfun.core.llms.vertexai import VertexAIClaude37Sonnet_20250219
from langfun.core.llms.vertexai import VertexAIClaude35Sonnet_20241022
from langfun.core.llms.vertexai import VertexAIClaude35Haiku_20241022
from langfun.core.llms.vertexai import VertexAIClaude3Opus_20240229

# Misc open source models.

# Gemma models.
from langfun.core.llms.groq import GroqGemma2_9B_IT

# Llama models.
from langfun.core.llms.vertexai import VertexAILlama
from langfun.core.llms.vertexai import VertexAILlama32_90B
from langfun.core.llms.vertexai import VertexAILlama31_405B
from langfun.core.llms.vertexai import VertexAILlama31_70B
from langfun.core.llms.vertexai import VertexAILlama31_8B

from langfun.core.llms.groq import GroqLlama33_70B_Versatile
from langfun.core.llms.groq import GroqLlama33_70B_SpecDec
from langfun.core.llms.groq import GroqLlama32_1B
from langfun.core.llms.groq import GroqLlama32_3B
from langfun.core.llms.groq import GroqLlama32_11B_Vision
from langfun.core.llms.groq import GroqLlama32_90B_Vision

# Mistral models.
from langfun.core.llms.vertexai import VertexAIMistral
from langfun.core.llms.vertexai import VertexAIMistralLarge_20241121
from langfun.core.llms.vertexai import VertexAICodestral_20250113

from langfun.core.llms.groq import GroqMistral_8x7B

# DeepSeek models.
from langfun.core.llms.deepseek import DeepSeek
from langfun.core.llms.deepseek import DeepSeekV3
from langfun.core.llms.deepseek import DeepSeekR1

# LLaMA C++ models.
from langfun.core.llms.llama_cpp import LlamaCppRemote

# Placeholder for Google-internal imports.

# Include cache as sub-module.
from langfun.core.llms import cache

# pylint: enable=g-import-not-at-top
# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Load balancers for environments."""

import abc
import threading

from langfun.env import interface
import pyglove as pg


class LoadBalancer(pg.Object):
  """Base class for load balancers."""

  @abc.abstractmethod
  def acquire(self, sandbox_pool: list[interface.Sandbox]) -> interface.Sandbox:
    """Acquires a free sandbox from a pool of sandboxes.

    The load balancer will pick a sandbox from the pool and mark it as pending.

    Args:
      sandbox_pool: The pool of sandboxes to pick from.

    Raises:
      IndexError: If all sandboxes in the pool are either busy or dead.
    """


class RoundRobin(LoadBalancer):
  """Round robin load balancer."""

  def _on_bound(self):
    super()._on_bound()
    self._counter = 0
    self._acquire_lock = threading.Lock()

  def acquire(self, sandbox_pool: list[interface.Sandbox]) -> interface.Sandbox:
    """Returns a free sandbox from the pool."""
    with self._acquire_lock:
      for _ in range(len(sandbox_pool)):
        sandbox = sandbox_pool[self._counter % len(sandbox_pool)]
        self._counter = self._counter + 1
        if sandbox.status == interface.Sandbox.Status.READY:
          # Mark the sandbox as acquired to prevent it from being acquired by
          # other threads.
          sandbox.set_acquired()
          return sandbox
    raise IndexError('No free sandbox in the pool.')



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JSON-based prompting protocol."""

import io
import textwrap
from typing import Any
from langfun.core.structured.schema import base
import pyglove as pg


class JsonError(Exception):  # pylint: disable=g-bad-exception-name
  """Json parsing error."""

  def __init__(self, json: str, cause: Exception):
    self.json = json
    self.cause = cause

  def __str__(self) -> str:
    r = io.StringIO()
    r.write(
        pg.colored(
            f'{self.cause.__class__.__name__}: {self.cause}', 'magenta'
        )
    )

    r.write('\n\n')
    r.write(pg.colored('JSON text:', 'red'))
    r.write('\n\n')
    r.write(textwrap.indent(pg.colored(self.json, 'magenta'), ' ' * 2))
    return r.getvalue()


class JsonPromptingProtocol(base.PromptingProtocol):
  """JSON-based prompting protocol."""

  NAME = 'json'

  def schema_repr(self, schema: base.Schema, **kwargs) -> str:
    del kwargs
    out = io.StringIO()
    def _visit(node: Any) -> None:
      if isinstance(node, str):
        out.write(f'"{node}"')
      elif isinstance(node, list):
        assert len(node) == 1, node
        out.write('[')
        _visit(node[0])
        out.write(']')
      elif isinstance(node, dict):
        out.write('{')
        for i, (k, v) in enumerate(node.items()):
          if i != 0:
            out.write(', ')
          out.write(f'"{k}": ')
          _visit(v)
        out.write('}')
      elif isinstance(node, pg.typing.Enum):
        out.write(' | '.join(
            f'"{v}"' if isinstance(v, str) else repr(v)
            for v in node.values))
      elif isinstance(node, pg.typing.PrimitiveType):
        x = node.value_type.__name__
        if isinstance(node, pg.typing.Number):
          params = []
          if node.min_value is not None:
            params.append(f'min={node.min_value}')
          if node.max_value is not None:
            params.append(f'max={node.max_value}')
          if params:
            x += f'({", ".join(params)})'
        elif isinstance(node, pg.typing.Str):
          if node.regex is not None:
            x += f'(regex={node.regex.pattern})'
        if node.is_noneable:
          x = x + ' | None'
        out.write(x)
      else:
        raise ValueError(
            f'Unsupported value spec being used as schema: {node}.')
    _visit(schema.schema_dict())
    return out.getvalue()

  def value_repr(
      self,
      value: Any,
      schema: base.Schema | None = None,
      **kwargs
  ) -> str:
    del schema, kwargs
    return pg.to_json_str(dict(result=value))

  def parse_value(
      self,
      text: str,
      schema: base.Schema | None = None,
      **kwargs
  ) -> Any:
    """Parses a JSON string into a structured object."""
    del schema
    try:
      text = cleanup_json(text)
      v = pg.from_json_str(text, **kwargs)
    except Exception as e:
      raise JsonError(text, e)  # pylint: disable=raise-missing-from

    if not isinstance(v, dict) or 'result' not in v:
      raise JsonError(text, ValueError(
          'The root node of the JSON must be a dict with key `result`. '
          f'Encountered: {v}'
      ))
    return v['result']


def cleanup_json(json_str: str) -> str:
  """Cleans up the LM responded JSON string."""
  # Treatments:
  # 1. Extract the JSON string with a top-level dict from the response.
  #    This prevents the leading and trailing texts in the response to
  #    be counted as part of the JSON.
  # 2. Escape new lines in JSON values.

  curly_brackets = 0
  under_json = False
  under_str = False
  str_begin = -1

  cleaned = io.StringIO()
  for i, c in enumerate(json_str):
    if c == '{' and not under_str:
      cleaned.write(c)
      curly_brackets += 1
      under_json = True
      continue
    elif not under_json:
      continue

    if c == '}' and not under_str:
      cleaned.write(c)
      curly_brackets -= 1
      if curly_brackets == 0:
        break
    elif c == '"' and json_str[i - 1] != '\\':
      under_str = not under_str
      if under_str:
        str_begin = i
      else:
        assert str_begin > 0
        str_value = json_str[str_begin : i + 1].replace('\n', '\\n')
        cleaned.write(str_value)
        str_begin = -1
    elif not under_str:
      cleaned.write(c)

  if not under_json:
    raise ValueError(f'No JSON dict in the output: {json_str}')

  if curly_brackets > 0:
    raise ValueError(
        f'Malformated JSON: missing {curly_brackets} closing curly braces.'
    )

  return cleaned.getvalue()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for base action."""

import asyncio
import time
import unittest

import langfun.core as lf
from langfun.core.agentic import action as action_lib
from langfun.core.llms import fake
import langfun.core.structured as lf_structured
import pyglove as pg


class Bar(action_lib.Action):
  simulate_action_error: bool = False
  simulate_execution_time: float = 0

  def call(self, session, *, lm, **kwargs):
    assert session.current_action.action is self
    session.info('Begin Bar')
    time.sleep(self.simulate_execution_time)
    session.query('bar', lm=lm)
    session.add_metadata(note='bar')
    session.update_progress('Query completed')
    if self.simulate_action_error:
      raise ValueError('Bar error')
    return 2 + pg.contextual_value('baz', 0)


class Foo(action_lib.Action):
  x: int
  simulate_action_error: bool = False
  simulate_query_error: bool = False
  simulate_execution_time: list[float] = [0, 0, 0, 0]
  max_bar_execution_time: float | None = None

  def call(self, session, *, lm, **kwargs):
    assert session.current_action.action is self
    with session.track_phase('prepare'):
      session.info('Begin Foo', x=1)
      time.sleep(self.simulate_execution_time[0])
      Bar()(session, lm=lm)
      session.query(
          'foo',
          schema=int if self.simulate_query_error else None,
          lm=lm
      )
    with session.track_queries():
      time.sleep(self.simulate_execution_time[1])
      self.make_additional_query(lm)
    session.add_metadata(note='foo')

    def _sub_task(i):
      session.add_metadata(**{f'subtask_{i}': i})
      time.sleep(self.simulate_execution_time[2])
      Bar()(session, lm=lm)
      return lf_structured.query(f'subtask_{i}', lm=lm)

    self._state = []
    for i, output, error in session.concurrent_map(
        _sub_task,
        range(3),
        max_workers=2,
        ordered=True,
        silence_on_errors=None,
    ):
      assert isinstance(i, int), i
      assert isinstance(output, str), output
      assert error is None, error
      self._state.append(i)
    return self.x + Bar(
        simulate_action_error=self.simulate_action_error,
        simulate_execution_time=self.simulate_execution_time[3]
    )(session, lm=lm, max_execution_time=self.max_bar_execution_time)

  def make_additional_query(self, lm):
    lf_structured.query('additional query', lm=lm)


class ExecutionUnitPositionTest(unittest.TestCase):

  def test_basics(self):
    pos1 = action_lib.ExecutionUnit.Position(None, 0)
    self.assertEqual(repr(pos1), 'Position(0)')
    self.assertEqual(str(pos1), '')
    self.assertIsNone(pos1.parent)
    self.assertEqual(pos1.index, 0)
    self.assertEqual(pos1.indices(), (0,))
    self.assertEqual(pos1, (0,))
    self.assertEqual(pos1, '')
    self.assertEqual(pos1, action_lib.ExecutionUnit.Position(None, 0))
    self.assertNotEqual(pos1, 1)
    self.assertNotEqual(pos1, (1,))
    self.assertNotEqual(pos1, action_lib.ExecutionUnit.Position(None, 1))

    pos2 = action_lib.ExecutionUnit.Position(pos1, 0)
    self.assertEqual(repr(pos2), 'Position(0, 0)')
    self.assertEqual(str(pos2), '1')
    self.assertEqual(pos2, '1')
    self.assertEqual(pos2.parent, pos1)
    self.assertEqual(pos2.index, 0)
    self.assertEqual(pos2.indices(), (0, 0))
    self.assertNotEqual(pos1, pos2)
    self.assertLess(pos1, pos2)
    self.assertGreater(pos2, pos1)
    self.assertEqual(
        hash(pos2),
        hash(
            action_lib.ExecutionUnit.Position(
                action_lib.ExecutionUnit.Position(None, 0), 0
            )
        )
    )

    pos3 = action_lib.ExecutionUnit.Position(pos2, 0)
    self.assertEqual(str(pos3), '1.1')
    self.assertEqual(pos3, '1.1')
    self.assertEqual(pos3.parent, pos2)
    self.assertEqual(pos3.index, 0)
    self.assertEqual(pos3.indices(), (0, 0, 0))
    self.assertEqual(pos3.to_str(separator='>'), '1>1')


class ActionInvocationTest(unittest.TestCase):

  def test_basics(self):
    action_invocation = action_lib.ActionInvocation(
        action=Foo(1)
    )
    self.assertEqual(action_invocation.id, '')
    root = action_lib.ActionInvocation(action=action_lib.RootAction())
    root.execution.append(action_invocation)
    self.assertIs(action_invocation.parent_action, root)
    self.assertEqual(action_invocation.id, '/a1')


class ExecutionTraceTest(unittest.TestCase):

  def test_basics(self):
    execution = action_lib.ExecutionTrace()
    self.assertEqual(execution.id, '')

    root = action_lib.ActionInvocation(action=action_lib.RootAction())
    action_invocation = action_lib.ActionInvocation(action=Foo(1))
    root.execution.append(action_invocation)
    self.assertEqual(action_invocation.execution.id, '/a1')

    root.execution.reset()
    self.assertFalse(root.execution)


class SessionTest(unittest.TestCase):

  def test_succeeded_trajectory(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    self.assertIsNone(foo.session)
    self.assertIsNone(foo.invocation)
    self.assertIsNone(foo.state)
    self.assertIsNone(foo.result)
    self.assertIsNone(foo.metadata)

    session = action_lib.Session(id='agent@1', verbose=True)
    self.assertEqual(session.id, 'agent@1')
    self.assertFalse(session.has_started)
    self.assertFalse(session.has_stopped)

    # Render HTML view to trigger dynamic update during execution.
    _ = session.to_html()

    with session:
      result = foo(session, lm=lm)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(result, 3)
    self.assertIsNone(foo.session)
    self.assertEqual(foo.state, [0, 1, 2])
    self.assertIs(foo.invocation.state, foo.state)
    self.assertEqual(foo.result, 3)
    self.assertEqual(
        foo.metadata, dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )

    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)
    self.assertIs(session.metadata, session.root.metadata)

    #
    # Inspecting the root invocation.
    #

    root = session.root
    self.assertIsNone(root.parent_action)
    self.assertEqual(root.id, 'agent@1:')
    self.assertEqual(root.execution.id, 'agent@1:')
    self.assertEqual(len(root.execution), 1)
    self.assertIs(root.execution[0].action, foo)

    self.assertTrue(root.execution.has_started)
    self.assertTrue(root.execution.has_stopped)
    self.assertGreater(root.execution.elapse, 0)
    self.assertEqual(root.result, 3)
    self.assertEqual(
        root.metadata,
        dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )

    # The root space should have one action (foo), no queries, and no logs.
    self.assertEqual(len(root.execution_units), 1)
    self.assertEqual(len(root.actions), 1)
    self.assertEqual(len(root.queries), 0)
    self.assertEqual(len(root.logs), 0)
    # 2 query from Bar, 2 from Foo and 2 * 3 from parallel executions.
    self.assertEqual(len(session.all_queries), 10)
    self.assertEqual(len(root.all_queries), 10)
    # 6 actions: Foo and 2 Bar, and 3 Bar from parallel executions.
    self.assertEqual(len(session.all_actions), 6)
    self.assertEqual(
        [str(a.position) for a in session.all_actions],
        ['1', '1.1', '1.2.1.1', '1.2.2.1', '1.2.3.1', '1.3']
    )
    self.assertEqual(len(root.all_actions), 6)
    # 1 log from Bar and 1 from Foo and 3 from Bar in parallel executions.
    self.assertEqual(len(session.all_logs), 6)
    self.assertEqual(len(root.all_logs), 6)
    self.assertIs(session.usage_summary, root.usage_summary)
    self.assertEqual(root.usage_summary.total.num_requests, 10)

    # Inspecting the top-level action (Foo)
    foo_invocation = root.execution[0]
    self.assertIs(foo.invocation, foo_invocation)
    self.assertIs(foo_invocation.parent_action, root)
    self.assertEqual(foo_invocation.id, 'agent@1:/a1')
    self.assertEqual(foo_invocation.execution.id, 'agent@1:/a1')
    self.assertEqual(len(foo_invocation.execution.items), 4)

    # Prepare phase.
    prepare_phase = foo_invocation.execution[0]
    self.assertIsNone(prepare_phase.position)
    self.assertIsInstance(prepare_phase, action_lib.ExecutionTrace)
    self.assertEqual(prepare_phase.id, 'agent@1:/a1/prepare')
    self.assertEqual(len(prepare_phase.items), 3)
    self.assertTrue(prepare_phase.has_started)
    self.assertTrue(prepare_phase.has_stopped)
    self.assertEqual(prepare_phase.usage_summary.total.num_requests, 2)
    self.assertIsInstance(prepare_phase.items[0], lf.logging.LogEntry)
    self.assertIsInstance(prepare_phase.items[1], action_lib.ActionInvocation)
    self.assertIs(prepare_phase.items[1].parent_execution_unit, foo_invocation)
    self.assertEqual(prepare_phase.items[1].id, 'agent@1:/a1/prepare/a1')
    self.assertIsInstance(prepare_phase.items[2], lf_structured.QueryInvocation)
    self.assertEqual(prepare_phase.items[2].id, 'agent@1:/a1/prepare/q1')

    # Tracked queries.
    query_invocation = foo_invocation.execution[1]
    self.assertIsInstance(query_invocation, lf_structured.QueryInvocation)
    self.assertEqual(query_invocation.id, 'agent@1:/a1/q2')
    self.assertIs(query_invocation.lm, lm)
    self.assertEqual(
        foo_invocation.execution.indexof(
            query_invocation, lf_structured.QueryInvocation
        ),
        1
    )
    self.assertEqual(
        root.execution.indexof(
            query_invocation, lf_structured.QueryInvocation
        ),
        -1
    )

    # Tracked parallel executions.
    parallel_executions = foo_invocation.execution[2]
    # root (0) > foo (0) > parallel executions (1)
    self.assertEqual(parallel_executions.position, (0, 0, 1))
    self.assertEqual(parallel_executions.id, 'agent@1:/a1/p1')
    self.assertIsInstance(parallel_executions, action_lib.ParallelExecutions)
    self.assertIs(
        parallel_executions.all_actions[0].parent_execution_unit,
        parallel_executions
    )
    self.assertIs(
        parallel_executions.all_actions[0].parent_action,
        foo_invocation
    )
    self.assertEqual(len(parallel_executions), 3)
    self.assertEqual(parallel_executions[0].id, 'agent@1:/a1/p1/b1')
    self.assertEqual(parallel_executions[1].id, 'agent@1:/a1/p1/b2')
    self.assertEqual(parallel_executions[2].id, 'agent@1:/a1/p1/b3')
    self.assertEqual(len(parallel_executions[0].execution_units), 1)
    self.assertEqual(len(parallel_executions[1].execution_units), 1)
    self.assertEqual(len(parallel_executions[2].execution_units), 1)
    self.assertEqual(len(parallel_executions[0].queries), 1)
    self.assertEqual(len(parallel_executions[0].all_queries), 2)
    self.assertEqual(len(parallel_executions[1].queries), 1)
    self.assertEqual(len(parallel_executions[1].all_queries), 2)
    self.assertEqual(len(parallel_executions[2].queries), 1)
    self.assertEqual(len(parallel_executions[2].all_queries), 2)
    self.assertEqual(len(parallel_executions.execution_units), 0)
    self.assertEqual(len(parallel_executions.actions), 0)
    self.assertEqual(len(parallel_executions.queries), 0)
    self.assertEqual(len(parallel_executions.logs), 0)
    self.assertEqual(len(parallel_executions.all_actions), 3)
    self.assertEqual(len(parallel_executions.all_queries), 6)
    self.assertEqual(len(parallel_executions.all_logs), 3)

    # Invocation to Bar.
    bar_invocation = foo_invocation.execution[3]
    self.assertIs(bar_invocation.parent_action, foo_invocation)
    self.assertIs(bar_invocation.parent_execution_unit, foo_invocation)
    self.assertEqual(bar_invocation.id, 'agent@1:/a1/a5')
    self.assertIsInstance(bar_invocation, action_lib.ActionInvocation)
    self.assertIsInstance(bar_invocation.action, Bar)
    self.assertEqual(bar_invocation.result, 2)
    self.assertEqual(bar_invocation.metadata, dict(note='bar'))
    self.assertEqual(len(bar_invocation.execution.items), 2)

    # Save to HTML
    self.assertIn('result', session.to_html().content)

    # Save session to JSON
    json_str = session.to_json_str(save_ref_value=True)
    self.assertIsInstance(pg.from_json_str(json_str), action_lib.Session)

  def test_acall(self):
    bar = Bar()
    with lf.context(baz=1):
      r = bar.acall(lm=fake.StaticResponse('lm response'))
      self.assertEqual(asyncio.run(r), 3)

  def test_failed_action(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      foo(lm=lm)

    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)

    # Inspecting the root invocation.
    root = session.root
    self.assertRegex(root.id, 'agent@.*:')
    self.assertTrue(root.has_error)
    foo_invocation = root.execution[0]
    self.assertIsInstance(foo_invocation, action_lib.ActionInvocation)
    self.assertTrue(foo_invocation.has_error)
    bar_invocation = foo_invocation.execution[3]
    self.assertIsInstance(bar_invocation, action_lib.ActionInvocation)
    self.assertTrue(bar_invocation.has_error)

    # Save to HTML
    self.assertIn('error', session.to_html().content)

  def test_failed_query(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_query_error=True)
    with self.assertRaisesRegex(lf_structured.MappingError, 'SyntaxError'):
      foo(lm=lm)

    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)

    # Inspecting the root invocation.
    root = session.root
    self.assertRegex(root.id, 'agent@.*:')
    self.assertTrue(root.has_error)
    foo_invocation = root.execution[0]
    self.assertIsInstance(foo_invocation, action_lib.ActionInvocation)
    self.assertTrue(foo_invocation.has_error)
    self.assertEqual(len(foo_invocation.execution.items), 3)

  def test_succeeded_with_implicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    foo(lm=lm, verbose=True)
    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)
    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(session.final_result, 3)
    self.assertFalse(session.root.has_error)
    self.assertEqual(session.root.metadata, {})

  def test_failed_with_implicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      foo(lm=lm)
    session = foo.session
    self.assertIsNotNone(session)
    self.assertIsInstance(session.root.action, action_lib.RootAction)
    self.assertIs(session.current_action, session.root)
    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertTrue(session.has_error)
    self.assertIsInstance(session.final_error, pg.ErrorInfo)
    self.assertIn('Bar error', str(session.root.error))

  def test_succeeded_with_explicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    self.assertIsNone(foo.session)
    self.assertIsNone(foo.result)
    self.assertIsNone(foo.metadata)

    session = action_lib.Session(id='agent@1')
    self.assertEqual(session.id, 'agent@1')
    self.assertFalse(session.has_started)
    self.assertFalse(session.has_stopped)

    with session:
      result = foo(session, lm=lm, verbose=True)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(result, 3)
    self.assertIsNone(foo.session)
    self.assertEqual(foo.result, 3)
    self.assertEqual(
        foo.metadata, dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )
    self.assertIs(session.final_result, foo.result)
    self.assertFalse(session.has_error)

  def test_succeeded_with_explicit_session_start_end(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1)
    self.assertIsNone(foo.session)
    self.assertIsNone(foo.result)
    self.assertIsNone(foo.metadata)

    session = action_lib.Session(id='agent@1')
    self.assertEqual(session.id, 'agent@1')
    self.assertFalse(session.has_started)
    self.assertFalse(session.has_stopped)

    session.start()
    result = foo(session, lm=lm)
    session.end(result)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(result, 3)
    self.assertIsNone(foo.session)
    self.assertEqual(foo.result, 3)
    self.assertEqual(
        foo.metadata, dict(note='foo', subtask_0=0, subtask_1=1, subtask_2=2)
    )
    self.assertIs(session.final_result, foo.result)
    self.assertFalse(session.has_error)

  def test_failed_with_explicit_session(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    session = action_lib.Session(id='agent@1')
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      with session:
        foo(session, lm=lm)
    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertTrue(session.has_error)
    self.assertIsNone(session.final_result)
    self.assertIsInstance(session.root.error, pg.ErrorInfo)
    self.assertIn('Bar error', str(session.root.error))

  def test_failed_with_explicit_session_without_start(self):
    lm = fake.StaticResponse('lm response')
    foo = Foo(1, simulate_action_error=True)
    session = action_lib.Session(id='agent@1')
    with self.assertRaisesRegex(ValueError, 'Please call `Session.start'):
      foo(session, lm=lm)

  def test_succeed_with_multiple_actions(self):
    lm = fake.StaticResponse('lm response')
    with action_lib.Session() as session:
      x = Bar()(session, lm=lm)
      y = Bar()(session, lm=lm)
      self.assertTrue(session.has_started)
      self.assertFalse(session.has_stopped)
      session.add_metadata(note='root metadata')
      session.end(x + y)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertEqual(session.final_result, 2 + 2)
    self.assertEqual(len(session.root.execution), 2)
    self.assertEqual(session.root.metadata, dict(note='root metadata'))

  def test_failed_with_multiple_actions(self):
    lm = fake.StaticResponse('lm response')
    with self.assertRaisesRegex(ValueError, 'Bar error'):
      with action_lib.Session() as session:
        x = Bar()(session, lm=lm)
        y = Bar(simulate_action_error=True)(session, lm=lm)
        session.end(x + y)

    self.assertTrue(session.has_started)
    self.assertTrue(session.has_stopped)
    self.assertTrue(session.has_error)
    self.assertIsInstance(session.root.error, pg.ErrorInfo)
    self.assertEqual(len(session.root.execution), 3)
    self.assertEqual(len(session.root.actions), 2)
    self.assertEqual(len(session.root.logs), 1)
    self.assertFalse(session.root.execution[0].has_error)
    self.assertTrue(session.root.execution[1].has_error)

  def test_max_execution_time(self):
    lm = fake.StaticResponse('lm response')
    bar = Bar(simulate_execution_time=1)
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Bar.*has exceeded .* 0.5 seconds'
    ):
      bar(lm=lm, max_execution_time=0.5)

    foo = Foo(1, simulate_execution_time=[0, 0, 0, 1])
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Foo.* has exceeded .* 0.5 seconds'
    ):
      foo(lm=lm, max_execution_time=0.5)

    # Timeout within concurrent_map.
    foo = Foo(1, simulate_execution_time=[0, 0, 1, 0])
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Foo.* has exceeded .* 0.5 seconds'
    ):
      foo(lm=lm, max_execution_time=0.5)

    # Timeout within bar.
    foo = Foo(
        1, simulate_execution_time=[0, 0, 0, 1], max_bar_execution_time=0.5
    )
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Bar.* has exceeded .* 0.5 seconds'
    ):
      foo(lm=lm)

    # Timeout within bar, however the effective max_execution_time of bar is the
    # remaining time of the parent action as it's smaller (0.5 < 1).
    foo = Foo(
        1, simulate_execution_time=[0, 0.5, 0, 1.0], max_bar_execution_time=1.0
    )
    with self.assertRaisesRegex(
        action_lib.ActionTimeoutError,
        'Action .*Foo.* has exceeded .*1.0 seconds'
    ):
      foo(lm=lm, max_execution_time=1.0)

  def test_event_handler(self):

    class MyActionHandler(pg.Object, action_lib.SessionEventHandler):
      def _on_bound(self):
        super()._on_bound()
        self.progresses = []

      def on_session_start(self, session):
        session.add_metadata(progresses=pg.Ref(self.progresses))

      def on_action_progress(self, session, action, title, **kwargs):
        self.progresses.append((action.id, title))

    handler = MyActionHandler()
    self.assertIs(handler.get(MyActionHandler), handler)
    self.assertIsNone(handler.get(action_lib.SessionLogging))

    handler_chain = action_lib.SessionEventHandlerChain(
        handlers=[handler, action_lib.SessionLogging()]
    )
    self.assertIs(handler_chain.get(MyActionHandler), handler)
    self.assertIs(
        handler_chain.get(action_lib.SessionLogging),
        handler_chain.handlers[1]
    )

    session = action_lib.Session(
        id='agent@1',
        event_handler=handler_chain
    )
    bar = Bar()
    with session:
      bar(session, lm=fake.StaticResponse('lm response'))
      session.update_progress('Trajectory completed')

    self.assertIs(session.metadata['progresses'], handler.progresses)
    self.assertEqual(handler.progresses, [
        ('agent@1:/a1', 'Query completed'),
        ('agent@1:', 'Trajectory completed'),
    ])

  def test_clone(self):
    event_handler = action_lib.SessionLogging()
    session = action_lib.Session(event_handler=event_handler)
    other = session.clone()
    self.assertIsNot(session, other)
    self.assertIs(other.event_handler, event_handler)

    other = session.clone(deep=True)
    self.assertIsNot(session, other)
    self.assertIsNot(other.event_handler, session.event_handler)

  def test_log(self):
    session = action_lib.Session()
    session.debug('hi', x=1, y=2)
    session.info('hi', x=1, y=2, for_action=session.root)
    session.warning('hi', x=1, y=2, for_action=session.root.action)
    session.error('hi', x=1, y=2)
    session.fatal('hi', x=1, y=2)

  def test_as_message(self):
    session = action_lib.Session()
    self.assertIn('agent@', session.id)
    self.assertIsInstance(session.as_message(), lf.AIMessage)

  def test_query_with_track_if(self):
    lm = fake.StaticResponse('lm response')
    session = action_lib.Session()

    # Render session to trigger javascript updates to the HTML when
    # operating on the session.
    _ = session.to_html()
    with session:
      # This query will succeed.
      session.query(
          'prompt1',
          schema=None,
          lm=lm,
          track_if=lambda q: not q.has_error,
          default=None)
      # This query will fail during parsing.
      session.query(
          'prompt2',
          schema=int,
          lm=lm,
          track_if=lambda q: not q.has_error,
          default=None)
    self.assertEqual(len(session.root.queries), 1)
    self.assertIsNone(session.root.queries[0].error)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checkpoint aggregator for Langfun evaluations."""

import concurrent.futures
import dataclasses
import os
import threading
import time
from typing import Annotated, Iterator

from langfun.core.eval.v2 import evaluation as evaluation_lib
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import reporting
from langfun.core.eval.v2.runners import base

import pyglove as pg


class CheckpointMonitor(base.RunnerBase):
  """Runner for monitoring checkpoing files generated by other runners.

  Currently checkpoint monitor only supports aggregating per-example
  checkpoint files.
  """

  NAME = 'checkpoint_monitor'

  plugins = [
      reporting.HtmlReporter(),
  ]

  checkpoint_pattern: Annotated[
      str, 'The glob pattern of the checkpoint files to monitor.'
  ] = 'checkpoint_*.bagz'

  monitor_inprogress_files: Annotated[
      bool,
      'If True, monitor in-progress files to aggregate.'
  ] = False

  poll_interval: Annotated[
      int,
      'The interval in seconds to poll for new checkpoint files.'
  ] = 5

  max_aggregation_threads: Annotated[
      int,
      'The maximum number of threads to aggregate checkpoints.'
  ] = 128

  bypass_old_ckpt_files_with_non_oop_errors: Annotated[
      bool,
      'If True, ignore old checkpoint files with non-oop errors.'
  ] = True

  ckpt_start_time: Annotated[
      float | None,
      (
          'The timestamp to treat checkpoint files modified before this '
          'time as old.'
      )
  ] = None

  @dataclasses.dataclass
  class _AggregationEntry:
    evaluation: evaluation_lib.Evaluation
    output_dir: str
    inprogress_file_pattern: str | None
    ckpt_file_pattern: str
    example_ids_inprogress: set[int]
    example_ids_to_be_aggregated: set[int]
    example_ids_being_aggregated: set[int]
    completion_lock: threading.Lock
    is_completed: bool = False

  def _on_bound(self):
    super()._on_bound()
    self._monitor_thread = None
    self._aggregation_entries = []
    self._aggregator_pool = None
    self._error = None
    if self.ckpt_start_time is None:
      self.rebind(ckpt_start_time=time.time(), skip_notification=True)
    self._ckpt_bypass_timestamp: dict[str, int] = {}

  def start(self):
    # Reset the experiment state before getting started.
    self.current_run.experiment.reset()

    # Signal the start of the run.
    self.on_run_start()

    # Start the non-leaf nodes.
    for node in self.current_run.experiment.nonleaf_nodes:
      self.on_experiment_start(node)

    for evaluation in self.current_run.experiment.leaf_nodes:
      # This is not precise, but we at least notify example start.
      if not self.current_run.filter or self.current_run.filter(evaluation):
        self.on_experiment_start(evaluation)

        # Signal the start of the examples if we are not monitoring in-progress
        # files.
        if not self.monitor_inprogress_files:
          for example_id in self.current_run.examples_to_evaluate(evaluation):
            self._mark_example_started(evaluation, example_id)

        # Create the aggregation entries for polling.
        output_dir = self.current_run.output_dir(evaluation)
        self._aggregation_entries.append(
            self._AggregationEntry(
                evaluation=evaluation,
                output_dir=output_dir,
                ckpt_file_pattern=os.path.join(
                    output_dir, self.checkpoint_pattern
                ),
                inprogress_file_pattern=os.path.join(
                    output_dir, '*.inprogress'
                ) if self.monitor_inprogress_files else None,
                example_ids_to_be_aggregated=(
                    self.current_run.examples_to_evaluate(evaluation)
                ),
                example_ids_inprogress=set(),
                example_ids_being_aggregated=set(),
                completion_lock=threading.Lock(),
                is_completed=False,
            )
        )
      else:
        self.on_experiment_skipped(evaluation)

    self._aggregator_pool = concurrent.futures.ThreadPoolExecutor(
        max_workers=self.max_aggregation_threads
    )
    self._monitor_thread = threading.Thread(target=self._monitor_loop)
    self._monitor_thread.start()

  def join(self):
    if self._monitor_thread:
      self._monitor_thread.join()
    if self._error is not None:
      raise self._error

  def run(self):
    self.start()
    self.join()

  def _monitor_loop(self):
    while not self._error and any(
        not e.is_completed for e in self._aggregation_entries
    ):
      for entry in self._aggregation_entries:
        if not entry.example_ids_to_be_aggregated:
          continue

        # Signal example processing.
        if self.monitor_inprogress_files:
          inprogress_files = pg.io.glob(entry.inprogress_file_pattern)
          for inprogress_file in inprogress_files:
            example_id = int(
                os.path.basename(inprogress_file).split('.')[0]
            )
            if example_id not in entry.example_ids_inprogress:
              self._mark_example_started(entry.evaluation, example_id)
              entry.example_ids_inprogress.add(example_id)

        for filepath in pg.io.glob(entry.ckpt_file_pattern):
          example_id = int(
              os.path.basename(filepath).split('.')[0].split('_')[-1]
          )
          if example_id in entry.example_ids_to_be_aggregated:
            last_modified_time = pg.io.getmtime(filepath)
            bypass_timestamp = self._ckpt_bypass_timestamp.get(filepath)
            if (
                bypass_timestamp is not None
                and last_modified_time <= bypass_timestamp
            ):
              continue

            # Remove example ID from the set to avoid duplicate processing.
            entry.example_ids_to_be_aggregated.remove(example_id)
            entry.example_ids_being_aggregated.add(example_id)

            # It could be that the example has been processed before, but the
            # inprogress file was removed. In this case, we should signal the
            # example has started before completing it.
            if example_id not in entry.example_ids_inprogress:
              self._mark_example_started(entry.evaluation, example_id)
              entry.example_ids_inprogress.add(example_id)

            self._aggregator_pool.submit(
                self._aggregate, entry, filepath, example_id, last_modified_time
            )
            pg.logging.info(
                '[%s] Aggregating example %d from %s...',
                entry.evaluation.id,
                example_id,
                filepath,
            )
      time.sleep(self.poll_interval)

    if self._error is None:
      self.on_run_complete()
    else:
      self.on_run_abort(self._error)

  def _aggregate(
      self,
      entry: _AggregationEntry,
      ckpt_filepath: str,
      example_id: int,
      last_modified_time: float,
  ):
    """Aggregate an example from a checkpoint file."""
    try:
      loaded_examples = entry.evaluation.state.load(
          ckpt_filepath,
          example_input_by_id=entry.evaluation.example_input_by_id,
          # Example metadata may be expensive to load, and is not used by
          # metric aggregation. Thus we do not load example metadata.
          load_example_metadata=False
      )
      assert len(loaded_examples) >= 1, loaded_examples
      # Ocassionally the per-example checkpoint file may contain the same
      # example processed multiple times. We only need to aggregate the last
      # example.
      example = loaded_examples[-1]
      if (
          self.bypass_old_ckpt_files_with_non_oop_errors
          and last_modified_time < self.ckpt_start_time
          and example.error is not None
          and not example.error.tag.startswith('MappingError')
      ):
        entry.example_ids_being_aggregated.remove(example_id)
        entry.example_ids_to_be_aggregated.add(example_id)
        self._ckpt_bypass_timestamp[ckpt_filepath] = last_modified_time
        pg.logging.info(
            '[%s] Bypassing old checkpoint file with non-oop errors (%s) '
            'for example %d, last_modified_time: %s, ckpt_start_time: %s',
            entry.evaluation.id,
            ckpt_filepath,
            example_id,
            last_modified_time,
            self.ckpt_start_time,
        )
        return
    except BaseException as e:  # pylint: disable=broad-except
      error_info = pg.ErrorInfo.from_exception(e)
      pg.logging.error(
          '[%s] Failed to aggregate example %d: %s',
          entry.evaluation.id,
          example_id,
          error_info
      )
      example = example_lib.Example(
          id=example_id,
          input=entry.evaluation.example_input_by_id(example_id),
          error=error_info,
      )

    # This will skip processing but still allow metrics to be collected.
    # `process` will never be called for evaluation, thus we do not
    # need to setup/teardown evaluation.
    try:
      example = entry.evaluation.evaluate(
          example, reevaluate_upon_previous_errors=False
      )
    except BaseException as e:  # pylint: disable=broad-except
      pg.logging.error(
          '[%s] Unexpected error found during evaluating example %d from %s.',
          entry.evaluation.id,
          example_id,
          ckpt_filepath,
      )
      self._error = e
      entry.example_ids_being_aggregated.remove(example_id)
      return

    example.newly_processed = True
    pg.logging.info(
        '[%s] Successfully aggregated example %d from %s.',
        entry.evaluation.id,
        example_id,
        ckpt_filepath,
    )

    try:
      self.on_example_complete(entry.evaluation, example)
    except BaseException as e:  # pylint: disable=broad-except
      # Plugin failures should be raised to the user.
      self._error = e

    entry.example_ids_being_aggregated.remove(example_id)

    # Remove the in-progress file to indicate that the example has been
    # processed.
    try:
      pg.io.rm(os.path.join(entry.output_dir, f'{example_id}.inprogress'))
    except FileNotFoundError:
      pass

    if (not self._error
        and not entry.example_ids_to_be_aggregated
        and not entry.example_ids_being_aggregated):
      with entry.completion_lock:
        if not entry.is_completed:
          entry.is_completed = True
          try:
            self.on_experiment_complete(entry.evaluation)
          except BaseException as e:  # pylint: disable=broad-except
            # Plugin failures should be raised to the user.
            self._error = e

  def _mark_example_started(
      self,
      evaluation: evaluation_lib.Evaluation,
      example_id: int
  ) -> None:
    """Mark an example as started."""
    example = example_lib.Example(
        id=example_id, input=evaluation.example_input_by_id(example_id),
    )
    example.start_time = time.time()
    self.on_example_start(evaluation, example)

    # We update evaluation state with the inprogress status so the evaluation
    # HTML could show remotely in-progress examples.
    evaluation.state.update(example, in_progress=True)

  def _run(self, evaluations: list[evaluation_lib.Evaluation]):
    raise NotImplementedError('Not needed in checkpoint monitor.')

  def _evaluate_items(
      self,
      evaluation: evaluation_lib.Evaluation,
      items: Iterator[example_lib.Example]
  ) -> None:
    raise NotImplementedError('Not needed in checkpoint monitor.')



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import unittest

from langfun.env import interface
from langfun.env import test_utils

TestingEnvironment = test_utils.TestingEnvironment
TestingSandbox = test_utils.TestingSandbox
TestingFeature = test_utils.TestingFeature
TestingEventHandler = test_utils.TestingEventHandler


class BaseEnvironmentTests(unittest.TestCase):

  def test_basics(self):
    env = TestingEnvironment(
        image_ids=['test_image'],
        root_dir='/tmp',
        pool_size=0,
        features={'test_feature': TestingFeature()},
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    self.assertIsNone(interface.Environment.current())
    self.assertEqual(env.image_ids, ['test_image'])
    self.assertFalse(env.supports_dynamic_image_loading)
    self.assertEqual(env.status, interface.Environment.Status.CREATED)
    self.assertFalse(env.is_online)
    self.assertEqual(env.min_pool_size('test_image'), 0)
    self.assertEqual(env.max_pool_size('test_image'), 0)
    self.assertEqual(env.sandbox_pool, {})
    self.assertEqual(env.id, interface.Environment.Id('testing-env'))
    self.assertEqual(env.outage_grace_period, 1)
    self.assertEqual(env.features['test_feature'].name, 'test_feature')

    self.assertIsNone(env.start_time)

    with env:
      self.assertEqual(env.status, interface.Environment.Status.ONLINE)
      self.assertIs(interface.Environment.current(), env)
      self.assertTrue(env.is_online)
      self.assertIsNotNone(env.start_time)
      self.assertEqual(env.offline_duration, 0.0)
      self.assertEqual(env.sandbox_pool, {})
      self.assertEqual(env.working_dir, '/tmp/testing-env')

      with env.sandbox('session1') as sb:
        self.assertEqual(
            sb.id, interface.Sandbox.Id(
                environment_id=env.id,
                image_id=sb.image_id,
                sandbox_id='0')
        )
        self.assertEqual(sb.session_id, 'session1')
        self.assertEqual(sb.working_dir, '/tmp/testing-env/test_image/0')
        self.assertTrue(sb.is_online)
        self.assertIs(sb.test_feature, sb.features['test_feature'])
        self.assertEqual(
            sb.test_feature.working_dir,
            '/tmp/testing-env/test_image/0/test_feature'
        )
        with self.assertRaises(AttributeError):
          _ = sb.test_feature2
      self.assertFalse(sb.is_online)

      with self.assertRaisesRegex(
          ValueError, 'Environment .* does not serve image ID .*'
      ):
        env.sandbox(image_id='test_image2')

      with env.test_feature() as feature:
        self.assertIsInstance(feature, TestingFeature)
        self.assertEqual(
            feature.sandbox.status, interface.Sandbox.Status.IN_SESSION
        )
        self.assertTrue(
            feature.sandbox.session_id.startswith('test_feature-session')
        )

      with self.assertRaises(AttributeError):
        _ = env.test_feature2

  def test_dynamic_image_loading(self):
    env = TestingEnvironment(
        image_ids=[],
        supports_dynamic_image_loading=True,
        pool_size=0,
        features={'test_feature': TestingFeature()},
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    with env:
      with env.sandbox(image_id='test_image2') as sb:
        self.assertEqual(sb.image_id, 'test_image2')

      with self.assertRaisesRegex(
          ValueError, 'Environment .* does not have a default image ID.'
      ):
        env.sandbox()

  def test_dynamic_image_loading_with_pooling(self):
    env = TestingEnvironment(
        image_ids=[],
        supports_dynamic_image_loading=True,
        pool_size=2,
        features={'test_feature': TestingFeature()},
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    with env:
      with env.sandbox(image_id='test_image'):
        self.assertEqual(len(env.sandbox_pool['test_image']), 1)

        with env.sandbox(image_id='test_image'):
          self.assertEqual(len(env.sandbox_pool['test_image']), 2)

          with self.assertRaises(interface.EnvironmentOverloadError):
            with env.sandbox(image_id='test_image'):
              pass
        self.assertEqual(len(env.sandbox_pool['test_image']), 2)

        with env.sandbox(image_id='test_image'):
          self.assertEqual(len(env.sandbox_pool['test_image']), 2)

  def test_image_feature_mappings(self):
    env = TestingEnvironment(
        image_ids=[
            'test_image1',
            'test_image2',
        ],
        features={
            'test_feature': TestingFeature(
                applicable_images=['test_image1.*']
            ),
            'test_feature2': TestingFeature(
                applicable_images=['test_image2.*']
            ),
            'test_feature3': TestingFeature(
                applicable_images=['test_image.*']
            ),
        },
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      with env.sandbox(image_id='test_image1') as sb:
        self.assertIn('test_feature', sb.features)
        self.assertNotIn('test_feature2', sb.features)
        self.assertIn('test_feature3', sb.features)

      with env.sandbox(image_id='test_image2') as sb:
        self.assertNotIn('test_feature', sb.features)
        self.assertIn('test_feature2', sb.features)
        self.assertIn('test_feature3', sb.features)

      with env.test_feature() as feature:
        self.assertEqual(feature.sandbox.image_id, 'test_image1')

      with self.assertRaisesRegex(
          ValueError, 'Feature .* is not applicable to .*'
      ):
        with env.test_feature(image_id='test_image2'):
          pass

      with env.test_feature2() as feature:
        self.assertEqual(feature.sandbox.image_id, 'test_image2')

      with env.test_feature3() as feature:
        self.assertEqual(feature.sandbox.image_id, 'test_image1')

      with env.test_feature3(image_id='test_image2') as feature:
        self.assertEqual(feature.sandbox.image_id, 'test_image2')

  def test_feature_applicability_check(self):
    with self.assertRaisesRegex(
        ValueError, 'Feature .* is not applicable to .*'
    ):
      TestingEnvironment(
          image_ids=[
              'test_image1',
          ],
          features={
              'test_feature2': TestingFeature(
                  applicable_images=['test_image2.*']
              ),
          },
      )
    env = TestingEnvironment(
        image_ids=[],
        supports_dynamic_image_loading=True,
        features={
            'test_feature2': TestingFeature(
                applicable_images=['test_image2.*']
            ),
        },
        pool_size=0
    )
    with env:
      with self.assertRaisesRegex(
          ValueError, 'No image ID found for feature .*'
      ):
        with env.test_feature2():
          pass

      # Dynamically loaded IDs.
      with env.test_feature2(image_id='test_image2') as feature:
        self.assertEqual(feature.sandbox.image_id, 'test_image2')

  def test_pool_size(self):
    env = TestingEnvironment(
        image_ids=['test_image'],
        pool_size=1,
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    self.assertEqual(env.min_pool_size('test_image'), 1)
    self.assertEqual(env.max_pool_size('test_image'), 1)

    env = TestingEnvironment(
        image_ids=['test_image'],
        pool_size=(0, 256),
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    self.assertEqual(env.min_pool_size('test_image'), 0)
    self.assertEqual(env.max_pool_size('test_image'), 256)

    env = TestingEnvironment(
        image_ids=['test_image'],
        pool_size={
            'test_.*': (0, 128),
            'my.*': (5, 64),
            'exact_image_name': 10,
        },
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    self.assertEqual(env.min_pool_size('test_image'), 0)
    self.assertEqual(env.max_pool_size('test_image'), 128)
    self.assertEqual(env.min_pool_size('my_image'), 5)
    self.assertEqual(env.max_pool_size('my_image'), 64)
    self.assertEqual(env.min_pool_size('exact_image_name'), 10)
    self.assertEqual(env.max_pool_size('exact_image_name'), 10)
    self.assertEqual(env.min_pool_size('some_image'), 0)  # default
    self.assertEqual(env.max_pool_size('some_image'), 256)  # default

  def test_del(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    env.start()
    sb = env.acquire()
    del sb
    del env

  def test_acquire_env_offline(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with self.assertRaises(interface.EnvironmentOutageError):
      env.acquire()

  def test_acquire_no_pooling(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)
      self.assertIsNone(env.working_dir)
      self.assertIsNone(sb.working_dir)
      self.assertIsNone(sb.test_feature.working_dir)

  def test_acquire_no_pooling_with_error(self):
    env = TestingEnvironment(
        features={
            'test_feature': TestingFeature(
                simulate_setup_error=interface.SandboxStateError
            )
        },
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      with self.assertRaises(interface.EnvironmentOutageError):
        env.acquire()

  def test_acquire_with_pooling(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=1,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)

  def test_acquire_with_pooling_overload(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=1,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)
      with self.assertRaises(interface.EnvironmentOverloadError):
        env.acquire()

  def test_acquire_with_growing_pool(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=(1, 3),
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      self.assertEqual(len(env.sandbox_pool['test_image']), 1)
      self.assertEqual(
          env.stats(),
          {
              'sandbox': {
                  'test_image': {
                      'created': 0,
                      'setting_up': 0,
                      'ready': 1,
                      'acquired': 0,
                      'in_session': 0,
                      'exiting_session': 0,
                      'shutting_down': 0,
                      'offline': 0,
                  }
              }
          }
      )
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)
      self.assertEqual(
          env.stats(),
          {
              'sandbox': {
                  'test_image': {
                      'created': 0,
                      'setting_up': 0,
                      'ready': 0,
                      'acquired': 1,
                      'in_session': 0,
                      'exiting_session': 0,
                      'shutting_down': 0,
                      'offline': 0,
                  }
              }
          }
      )
      self.assertEqual(len(env.sandbox_pool['test_image']), 1)
      sb2 = env.acquire()
      self.assertEqual(sb2.status, interface.Sandbox.Status.ACQUIRED)
      self.assertEqual(len(env.sandbox_pool['test_image']), 2)
      self.assertEqual(
          env.stats(),
          {
              'sandbox': {
                  'test_image': {
                      'created': 0,
                      'setting_up': 0,
                      'ready': 0,
                      'acquired': 2,
                      'in_session': 0,
                      'exiting_session': 0,
                      'shutting_down': 0,
                      'offline': 0,
                  }
              }
          }
      )
    self.assertEqual(
        env.stats(),
        {
            'sandbox': {}
        }
    )

  def test_acquire_with_growing_pool_failure(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=(1, 3),
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      self.assertEqual(len(env.sandbox_pool), 1)
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)

      # Make future sandbox setup to fail.
      env.features.test_feature.rebind(
          simulate_setup_error=interface.SandboxStateError,
          skip_notification=True
      )
      with self.assertRaises(interface.EnvironmentOutageError):
        env.acquire()

  def test_housekeep_error(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=1,
        proactive_session_setup=True,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      self.assertEqual(len(env.sandbox_pool), 1)
      self.assertIn('test_image', env.sandbox_pool)
      self.assertEqual(
          env.sandbox_pool['test_image'][0].status,
          interface.Sandbox.Status.READY
      )
      # Make future sandbox setup to fail.
      env.features.test_feature.rebind(
          simulate_setup_error=interface.SandboxStateError,
          skip_notification=True
      )
      with self.assertRaises(interface.SandboxStateError):
        with env.sandbox() as sb:
          sb.shell('bad command', raise_error=interface.SandboxStateError)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(len(sb.state_errors), 1)
      sb_offline_time = time.time()
      while time.time() - sb_offline_time < 10:
        if not env.is_online:
          break
        time.sleep(0.5)
      self.assertFalse(env.is_online)


if __name__ == '__main__':
  unittest.main()
